#!/usr/bin/env python3

import argparse
import copy
import json
import logging
import os
import pathlib
import pickle
import re
import time
from datetime import UTC, datetime
from importlib.metadata import version

import numpy as np
import pandas as pd

from emhass import utils
from emhass.forecast import Forecast
from emhass.machine_learning_forecaster import MLForecaster
from emhass.machine_learning_regressor import MLRegressor
from emhass.optimization import Optimization
from emhass.retrieve_hass import RetrieveHass

default_csv_filename = "opt_res_latest.csv"
default_pkl_suffix = "_mlf.pkl"
default_metadata_json = "metadata.json"


def retrieve_home_assistant_data(
    set_type: str,
    get_data_from_file: bool,
    retrieve_hass_conf: dict,
    optim_conf: dict,
    rh: RetrieveHass,
    emhass_conf: dict,
    test_df_literal: pd.DataFrame,
) -> dict:
    """Retrieve data from Home Assistant or file and prepare it for optimization."""
    if get_data_from_file:
        with open(emhass_conf["data_path"] / test_df_literal, "rb") as inp:
            rh.df_final, days_list, var_list, rh.ha_config = pickle.load(inp)
            rh.var_list = var_list
        # Assign variables based on set_type
        retrieve_hass_conf["sensor_power_load_no_var_loads"] = str(var_list[0])
        if optim_conf.get("set_use_pv", True):
            retrieve_hass_conf["sensor_power_photovoltaics"] = str(var_list[1])
            retrieve_hass_conf["sensor_linear_interp"] = [
                retrieve_hass_conf["sensor_power_photovoltaics"],
                retrieve_hass_conf["sensor_power_load_no_var_loads"],
            ]
            retrieve_hass_conf["sensor_replace_zero"] = [
                retrieve_hass_conf["sensor_power_photovoltaics"],
                var_list[2],
            ]
        else:
            retrieve_hass_conf["sensor_linear_interp"] = [
                retrieve_hass_conf["sensor_power_load_no_var_loads"]
            ]
            retrieve_hass_conf["sensor_replace_zero"] = []
    else:
        # Determine days_list based on set_type
        if set_type == "perfect-optim" or set_type == "adjust_pv":
            days_list = utils.get_days_list(
                retrieve_hass_conf["historic_days_to_retrieve"]
            )
        elif set_type == "naive-mpc-optim":
            days_list = utils.get_days_list(1)
        else:
            days_list = None  # Not needed for dayahead
        var_list = [retrieve_hass_conf["sensor_power_load_no_var_loads"]]
        if optim_conf.get("set_use_pv", True):
            var_list.append(retrieve_hass_conf["sensor_power_photovoltaics"])
            if optim_conf.get("set_use_adjusted_pv", True):
                var_list.append(
                    retrieve_hass_conf["sensor_power_photovoltaics_forecast"]
                )
        if not rh.get_data(
            days_list, var_list, minimal_response=False, significant_changes_only=False
        ):
            return False, None, days_list
    rh.prepare_data(
        retrieve_hass_conf["sensor_power_load_no_var_loads"],
        load_negative=retrieve_hass_conf["load_negative"],
        set_zero_min=retrieve_hass_conf["set_zero_min"],
        var_replace_zero=retrieve_hass_conf["sensor_replace_zero"],
        var_interp=retrieve_hass_conf["sensor_linear_interp"],
    )
    return True, rh.df_final.copy(), days_list


def adjust_pv_forecast(
    logger: logging.Logger,
    fcst: Forecast,
    P_PV_forecast: pd.Series,
    get_data_from_file: bool,
    retrieve_hass_conf: dict,
    optim_conf: dict,
    rh: RetrieveHass,
    emhass_conf: dict,
    test_df_literal: pd.DataFrame,
) -> pd.Series:
    """
    Adjust the photovoltaic (PV) forecast using historical data and a regression model.

    This method retrieves historical data, prepares it for model fitting, trains a regression
    model, and adjusts the provided PV forecast based on the trained model.

    :param logger: Logger object for logging information and errors.
    :type logger: logging.Logger
    :param fcst: Forecast object used for PV forecast adjustment.
    :type fcst: Forecast
    :param P_PV_forecast: The initial PV forecast to be adjusted.
    :type P_PV_forecast: pd.Series
    :param get_data_from_file: Whether to retrieve data from a file instead of Home Assistant.
    :type get_data_from_file: bool
    :param retrieve_hass_conf: Configuration dictionary for retrieving data from Home Assistant.
    :type retrieve_hass_conf: dict
    :param optim_conf: Configuration dictionary for optimization settings.
    :type optim_conf: dict
    :param rh: RetrieveHass object for interacting with Home Assistant.
    :type rh: RetrieveHass
    :param emhass_conf: Configuration dictionary for emhass paths and settings.
    :type emhass_conf: dict
    :param test_df_literal: DataFrame containing test data for debugging purposes.
    :type test_df_literal: pd.DataFrame
    :return: The adjusted PV forecast as a pandas Series.
    :rtype: pd.Series
    """
    logger.info("Adjusting PV forecast, retrieving history data for model fit")
    # Retrieve data from Home Assistant
    success, df_input_data, _ = retrieve_home_assistant_data(
        "adjust_pv",
        get_data_from_file,
        retrieve_hass_conf,
        optim_conf,
        rh,
        emhass_conf,
        test_df_literal,
    )
    if not success:
        return False
    # Call data preparation method
    fcst.adjust_pv_forecast_data_prep(df_input_data)
    # Call the fit method
    fcst.adjust_pv_forecast_fit(
        n_splits=5,
        regression_model=optim_conf["adjusted_pv_regression_model"],
    )
    # Call the predict method
    P_PV_forecast = P_PV_forecast.rename("forecast").to_frame()
    P_PV_forecast = fcst.adjust_pv_forecast_predict(forecasted_pv=P_PV_forecast)
    # Update the PV forecast
    return P_PV_forecast["adjusted_forecast"].rename(None)


def set_input_data_dict(
    emhass_conf: dict,
    costfun: str,
    params: str,
    runtimeparams: str,
    set_type: str,
    logger: logging.Logger,
    get_data_from_file: bool | None = False,
) -> dict:
    """
    Set up some of the data needed for the different actions.

    :param emhass_conf: Dictionary containing the needed emhass paths
    :type emhass_conf: dict
    :param costfun: The type of cost function to use for optimization problem
    :type costfun: str
    :param params: Configuration parameters passed from data/options.json
    :type params: str
    :param runtimeparams: Runtime optimization parameters passed as a dictionary
    :type runtimeparams: str
    :param set_type: Set the type of setup based on following type of optimization
    :type set_type: str
    :param logger: The passed logger object
    :type logger: logging object
    :param get_data_from_file: Use data from saved CSV file (useful for debug)
    :type get_data_from_file: bool, optional
    :return: A dictionnary with multiple data used by the action functions
    :rtype: dict

    """
    logger.info("Setting up needed data")

    # check if passed params is a dict
    if (params is not None) and (params != "null"):
        if type(params) is str:
            params = json.loads(params)
    else:
        params = {}

    # Parsing yaml
    retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(params, logger)
    if type(retrieve_hass_conf) is bool:
        return False

    # Treat runtimeparams
    params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
        runtimeparams,
        params,
        retrieve_hass_conf,
        optim_conf,
        plant_conf,
        set_type,
        logger,
        emhass_conf,
    )

    # Define the data retrieve object
    rh = RetrieveHass(
        retrieve_hass_conf["hass_url"],
        retrieve_hass_conf["long_lived_token"],
        retrieve_hass_conf["optimization_time_step"],
        retrieve_hass_conf["time_zone"],
        params,
        emhass_conf,
        logger,
        get_data_from_file=get_data_from_file,
    )

    # Retrieve basic configuration data from hass
    test_df_literal = "test_df_final.pkl"
    if get_data_from_file:
        with open(emhass_conf["data_path"] / test_df_literal, "rb") as inp:
            _, _, _, rh.ha_config = pickle.load(inp)
    else:
        response = rh.get_ha_config()
        if type(response) is bool:
            return False

    # Update the params dict using data from the HA configuration
    params = utils.update_params_with_ha_config(
        params,
        rh.ha_config,
    )

    # Define the forecast and optimization objects
    fcst = Forecast(
        retrieve_hass_conf,
        optim_conf,
        plant_conf,
        params,
        emhass_conf,
        logger,
        get_data_from_file=get_data_from_file,
    )
    opt = Optimization(
        retrieve_hass_conf,
        optim_conf,
        plant_conf,
        fcst.var_load_cost,
        fcst.var_prod_price,
        costfun,
        emhass_conf,
        logger,
    )

    # Perform setup based on type of action
    if set_type == "perfect-optim":
        # Retrieve data from hass
        success, df_input_data, days_list = retrieve_home_assistant_data(
            set_type,
            get_data_from_file,
            retrieve_hass_conf,
            optim_conf,
            rh,
            emhass_conf,
            test_df_literal,
        )
        if not success:
            return False
        # What we don't need for this type of action
        P_PV_forecast, P_load_forecast, df_input_data_dayahead = None, None, None
    elif set_type == "dayahead-optim":
        # Get PV and load forecasts
        if (
            optim_conf["set_use_pv"]
            or optim_conf.get("weather_forecast_method", None) == "list"
        ):
            df_weather = fcst.get_weather_forecast(
                method=optim_conf["weather_forecast_method"]
            )
            if isinstance(df_weather, bool) and not df_weather:
                return False
            P_PV_forecast = fcst.get_power_from_weather(df_weather)
            # Adjust PV forecast
            if optim_conf["set_use_adjusted_pv"]:
                # Update the PV forecast
                P_PV_forecast = adjust_pv_forecast(
                    logger,
                    fcst,
                    P_PV_forecast,
                    get_data_from_file,
                    retrieve_hass_conf,
                    optim_conf,
                    rh,
                    emhass_conf,
                    test_df_literal,
                )
        else:
            P_PV_forecast = pd.Series(0, index=fcst.forecast_dates)
        P_load_forecast = fcst.get_load_forecast(
            days_min_load_forecast=optim_conf["delta_forecast_daily"].days,
            method=optim_conf["load_forecast_method"],
        )
        if isinstance(P_load_forecast, bool) and not P_load_forecast:
            logger.error(
                "Unable to get sensor power photovoltaics, or sensor power load no var loads. Check HA sensors and their daily data"
            )
            return False
        df_input_data_dayahead = pd.DataFrame(
            np.transpose(np.vstack([P_PV_forecast.values, P_load_forecast.values])),
            index=P_PV_forecast.index,
            columns=["P_PV_forecast", "P_load_forecast"],
        )
        if (
            "optimization_time_step" in retrieve_hass_conf
            and retrieve_hass_conf["optimization_time_step"]
        ):
            if not isinstance(
                retrieve_hass_conf["optimization_time_step"],
                pd._libs.tslibs.timedeltas.Timedelta,
            ):
                optimization_time_step = pd.to_timedelta(
                    retrieve_hass_conf["optimization_time_step"], "minute"
                )
            else:
                optimization_time_step = retrieve_hass_conf["optimization_time_step"]
            df_input_data_dayahead = df_input_data_dayahead.asfreq(
                optimization_time_step
            )
        else:
            df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)
        params = json.loads(params)
        if (
            "prediction_horizon" in params["passed_data"]
            and params["passed_data"]["prediction_horizon"] is not None
        ):
            prediction_horizon = params["passed_data"]["prediction_horizon"]
            df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[
                df_input_data_dayahead.index[0] : df_input_data_dayahead.index[
                    prediction_horizon - 1
                ]
            ]
        # What we don't need for this type of action
        df_input_data, days_list = None, None
    elif set_type == "naive-mpc-optim":
        if (
            optim_conf.get("load_forecast_method", None) == "list"
            and optim_conf.get("weather_forecast_method", None) == "list"
        ) or (
            optim_conf.get("load_forecast_method", None) == "list"
            and not (optim_conf["set_use_pv"])
        ):
            days_list = None
            set_mix_forecast = False
            df_input_data = None
        else:
            # Retrieve data from hass
            success, df_input_data, days_list = retrieve_home_assistant_data(
                set_type,
                get_data_from_file,
                retrieve_hass_conf,
                optim_conf,
                rh,
                emhass_conf,
                test_df_literal,
            )
            if not success:
                return False
            set_mix_forecast = True
        # Get PV and load forecasts
        if (
            optim_conf["set_use_pv"]
            or optim_conf.get("weather_forecast_method", None) == "list"
        ):
            df_weather = fcst.get_weather_forecast(
                method=optim_conf["weather_forecast_method"]
            )
            if isinstance(df_weather, bool) and not df_weather:
                return False
            P_PV_forecast = fcst.get_power_from_weather(
                df_weather, set_mix_forecast=set_mix_forecast, df_now=df_input_data
            )
            # Adjust PV forecast
            if optim_conf["set_use_adjusted_pv"]:
                # Update the PV forecast
                P_PV_forecast = adjust_pv_forecast(
                    logger,
                    fcst,
                    P_PV_forecast,
                    get_data_from_file,
                    retrieve_hass_conf,
                    optim_conf,
                    rh,
                    emhass_conf,
                    test_df_literal,
                )
        else:
            P_PV_forecast = pd.Series(0, index=fcst.forecast_dates)
        P_load_forecast = fcst.get_load_forecast(
            days_min_load_forecast=optim_conf["delta_forecast_daily"].days,
            method=optim_conf["load_forecast_method"],
            set_mix_forecast=set_mix_forecast,
            df_now=df_input_data,
        )
        if isinstance(P_load_forecast, bool) and not P_load_forecast:
            logger.error(
                "Unable to get sensor power photovoltaics, or sensor power load no var loads. Check HA sensors and their daily data"
            )
            return False
        df_input_data_dayahead = pd.concat([P_PV_forecast, P_load_forecast], axis=1)
        if (
            "optimization_time_step" in retrieve_hass_conf
            and retrieve_hass_conf["optimization_time_step"]
        ):
            if not isinstance(
                retrieve_hass_conf["optimization_time_step"],
                pd._libs.tslibs.timedeltas.Timedelta,
            ):
                optimization_time_step = pd.to_timedelta(
                    retrieve_hass_conf["optimization_time_step"], "minute"
                )
            else:
                optimization_time_step = retrieve_hass_conf["optimization_time_step"]
            df_input_data_dayahead = df_input_data_dayahead.asfreq(
                optimization_time_step
            )
        else:
            df_input_data_dayahead = utils.set_df_index_freq(df_input_data_dayahead)
        df_input_data_dayahead.columns = ["P_PV_forecast", "P_load_forecast"]
        params = json.loads(params)
        if (
            "prediction_horizon" in params["passed_data"]
            and params["passed_data"]["prediction_horizon"] is not None
        ):
            prediction_horizon = params["passed_data"]["prediction_horizon"]
            df_input_data_dayahead = copy.deepcopy(df_input_data_dayahead)[
                df_input_data_dayahead.index[0] : df_input_data_dayahead.index[
                    prediction_horizon - 1
                ]
            ]
    elif (
        set_type == "forecast-model-fit"
        or set_type == "forecast-model-predict"
        or set_type == "forecast-model-tune"
    ):
        df_input_data_dayahead = None
        P_PV_forecast, P_load_forecast = None, None
        params = json.loads(params)
        # Retrieve data from hass
        days_to_retrieve = params["passed_data"]["historic_days_to_retrieve"]
        model_type = params["passed_data"]["model_type"]
        var_model = params["passed_data"]["var_model"]
        if get_data_from_file:
            days_list = None
            filename = model_type + ".pkl"
            filename_path = emhass_conf["data_path"] / filename
            with open(filename_path, "rb") as inp:
                df_input_data, _, _, _ = pickle.load(inp)
            df_input_data = df_input_data[
                df_input_data.index[-1] - pd.offsets.Day(days_to_retrieve) :
            ]
        else:
            days_list = utils.get_days_list(days_to_retrieve)
            var_list = [var_model]
            if not rh.get_data(days_list, var_list):
                return False
            df_input_data = rh.df_final.copy()
    elif set_type == "regressor-model-fit" or set_type == "regressor-model-predict":
        df_input_data, df_input_data_dayahead = None, None
        P_PV_forecast, P_load_forecast = None, None
        params = json.loads(params)
        days_list = None
        csv_file = params["passed_data"].get("csv_file", None)
        if "features" in params["passed_data"]:
            features = params["passed_data"]["features"]
        if "target" in params["passed_data"]:
            target = params["passed_data"]["target"]
        if "timestamp" in params["passed_data"]:
            timestamp = params["passed_data"]["timestamp"]
        if csv_file:
            if get_data_from_file:
                base_path = emhass_conf["data_path"]  # + "/data"
                filename_path = pathlib.Path(base_path) / csv_file
            else:
                filename_path = emhass_conf["data_path"] / csv_file
            if filename_path.is_file():
                df_input_data = pd.read_csv(filename_path, parse_dates=True)
            else:
                logger.error(
                    "The CSV file "
                    + csv_file
                    + " was not found in path: "
                    + str(emhass_conf["data_path"])
                )
                return False
            required_columns = []
            required_columns.extend(features)
            required_columns.append(target)
            if timestamp is not None:
                required_columns.append(timestamp)
            if not set(required_columns).issubset(df_input_data.columns):
                logger.error("The cvs file does not contain the required columns.")
                msg = f"CSV file should contain the following columns: {', '.join(required_columns)}"
                logger.error(msg)
                return False
    elif set_type == "publish-data":
        df_input_data, df_input_data_dayahead = None, None
        P_PV_forecast, P_load_forecast = None, None
        days_list = None
    else:
        logger.error(
            "The passed action argument and hence the set_type parameter for setup is not valid",
        )
        df_input_data, df_input_data_dayahead = None, None
        P_PV_forecast, P_load_forecast = None, None
        days_list = None
    # The input data dictionary to return
    input_data_dict = {
        "emhass_conf": emhass_conf,
        "retrieve_hass_conf": retrieve_hass_conf,
        "rh": rh,
        "opt": opt,
        "fcst": fcst,
        "df_input_data": df_input_data,
        "df_input_data_dayahead": df_input_data_dayahead,
        "P_PV_forecast": P_PV_forecast,
        "P_load_forecast": P_load_forecast,
        "costfun": costfun,
        "params": params,
        "days_list": days_list,
    }
    return input_data_dict


def weather_forecast_cache(
    emhass_conf: dict, params: str, runtimeparams: str, logger: logging.Logger
) -> bool:
    """
    Perform a call to get forecast function, intend to save results to cache.

    :param emhass_conf: Dictionary containing the needed emhass paths
    :type emhass_conf: dict
    :param params: Configuration parameters passed from data/options.json
    :type params: str
    :param runtimeparams: Runtime optimization parameters passed as a dictionary
    :type runtimeparams: str
    :param logger: The passed logger object
    :type logger: logging object
    :return: A bool for function completion
    :rtype: bool

    """
    # Parsing yaml
    retrieve_hass_conf, optim_conf, plant_conf = utils.get_yaml_parse(params, logger)
    # Treat runtimeparams
    params, retrieve_hass_conf, optim_conf, plant_conf = utils.treat_runtimeparams(
        runtimeparams,
        params,
        retrieve_hass_conf,
        optim_conf,
        plant_conf,
        "forecast",
        logger,
        emhass_conf,
    )
    # Make sure weather_forecast_cache is true
    if (params is not None) and (params != "null"):
        params = json.loads(params)
    else:
        params = {}
    params["passed_data"]["weather_forecast_cache"] = True
    params = json.dumps(params)
    # Create Forecast object
    fcst = Forecast(
        retrieve_hass_conf, optim_conf, plant_conf, params, emhass_conf, logger
    )
    result = fcst.get_weather_forecast(optim_conf["weather_forecast_method"])
    if isinstance(result, bool) and not result:
        return False

    return True


def perfect_forecast_optim(
    input_data_dict: dict,
    logger: logging.Logger,
    save_data_to_file: bool | None = True,
    debug: bool | None = False,
) -> pd.DataFrame:
    """
    Perform a call to the perfect forecast optimization routine.

    :param input_data_dict:  A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging object
    :param save_data_to_file: Save optimization results to CSV file
    :type save_data_to_file: bool, optional
    :param debug: A debug option useful for unittests
    :type debug: bool, optional
    :return: The output data of the optimization
    :rtype: pd.DataFrame

    """
    logger.info("Performing perfect forecast optimization")
    # Load cost and prod price forecast
    df_input_data = input_data_dict["fcst"].get_load_cost_forecast(
        input_data_dict["df_input_data"],
        method=input_data_dict["fcst"].optim_conf["load_cost_forecast_method"],
        list_and_perfect=True,
    )
    if isinstance(df_input_data, bool) and not df_input_data:
        return False
    df_input_data = input_data_dict["fcst"].get_prod_price_forecast(
        df_input_data,
        method=input_data_dict["fcst"].optim_conf["production_price_forecast_method"],
        list_and_perfect=True,
    )
    if isinstance(df_input_data, bool) and not df_input_data:
        return False
    opt_res = input_data_dict["opt"].perform_perfect_forecast_optim(
        df_input_data, input_data_dict["days_list"]
    )
    # Save CSV file for analysis
    if save_data_to_file:
        filename = "opt_res_perfect_optim_" + input_data_dict["costfun"] + ".csv"
    else:  # Just save the latest optimization results
        filename = default_csv_filename
    if not debug:
        opt_res.to_csv(
            input_data_dict["emhass_conf"]["data_path"] / filename,
            index_label="timestamp",
        )
    if not isinstance(input_data_dict["params"], dict):
        params = json.loads(input_data_dict["params"])
    else:
        params = input_data_dict["params"]

    # if continual_publish, save perfect results to data_path/entities json
    if input_data_dict["retrieve_hass_conf"].get("continual_publish", False) or params[
        "passed_data"
    ].get("entity_save", False):
        # Trigger the publish function, save entity data and not post to HA
        publish_data(input_data_dict, logger, entity_save=True, dont_post=True)

    return opt_res


def dayahead_forecast_optim(
    input_data_dict: dict,
    logger: logging.Logger,
    save_data_to_file: bool | None = False,
    debug: bool | None = False,
) -> pd.DataFrame:
    """
    Perform a call to the day-ahead optimization routine.

    :param input_data_dict:  A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging object
    :param save_data_to_file: Save optimization results to CSV file
    :type save_data_to_file: bool, optional
    :param debug: A debug option useful for unittests
    :type debug: bool, optional
    :return: The output data of the optimization
    :rtype: pd.DataFrame

    """
    logger.info("Performing day-ahead forecast optimization")
    # Load cost and prod price forecast
    df_input_data_dayahead = input_data_dict["fcst"].get_load_cost_forecast(
        input_data_dict["df_input_data_dayahead"],
        method=input_data_dict["fcst"].optim_conf["load_cost_forecast_method"],
    )
    if isinstance(df_input_data_dayahead, bool) and not df_input_data_dayahead:
        return False
    df_input_data_dayahead = input_data_dict["fcst"].get_prod_price_forecast(
        df_input_data_dayahead,
        method=input_data_dict["fcst"].optim_conf["production_price_forecast_method"],
    )
    if isinstance(df_input_data_dayahead, bool) and not df_input_data_dayahead:
        return False
    if "outdoor_temperature_forecast" in input_data_dict["params"]["passed_data"]:
        df_input_data_dayahead["outdoor_temperature_forecast"] = input_data_dict[
            "params"
        ]["passed_data"]["outdoor_temperature_forecast"]
    opt_res_dayahead = input_data_dict["opt"].perform_dayahead_forecast_optim(
        df_input_data_dayahead,
        input_data_dict["P_PV_forecast"],
        input_data_dict["P_load_forecast"],
    )
    # Save CSV file for publish_data
    if save_data_to_file:
        today = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
        filename = "opt_res_dayahead_" + today.strftime("%Y_%m_%d") + ".csv"
    else:  # Just save the latest optimization results
        filename = default_csv_filename
    if not debug:
        opt_res_dayahead.to_csv(
            input_data_dict["emhass_conf"]["data_path"] / filename,
            index_label="timestamp",
        )

    if not isinstance(input_data_dict["params"], dict):
        params = json.loads(input_data_dict["params"])
    else:
        params = input_data_dict["params"]

    # if continual_publish, save day_ahead results to data_path/entities json
    if input_data_dict["retrieve_hass_conf"].get("continual_publish", False) or params[
        "passed_data"
    ].get("entity_save", False):
        # Trigger the publish function, save entity data and not post to HA
        publish_data(input_data_dict, logger, entity_save=True, dont_post=True)

    return opt_res_dayahead


def naive_mpc_optim(
    input_data_dict: dict,
    logger: logging.Logger,
    save_data_to_file: bool | None = False,
    debug: bool | None = False,
) -> pd.DataFrame:
    """
    Perform a call to the naive Model Predictive Controller optimization routine.

    :param input_data_dict:  A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging object
    :param save_data_to_file: Save optimization results to CSV file
    :type save_data_to_file: bool, optional
    :param debug: A debug option useful for unittests
    :type debug: bool, optional
    :return: The output data of the optimization
    :rtype: pd.DataFrame

    """
    logger.info("Performing naive MPC optimization")
    # Load cost and prod price forecast
    df_input_data_dayahead = input_data_dict["fcst"].get_load_cost_forecast(
        input_data_dict["df_input_data_dayahead"],
        method=input_data_dict["fcst"].optim_conf["load_cost_forecast_method"],
    )
    if isinstance(df_input_data_dayahead, bool) and not df_input_data_dayahead:
        return False
    df_input_data_dayahead = input_data_dict["fcst"].get_prod_price_forecast(
        df_input_data_dayahead,
        method=input_data_dict["fcst"].optim_conf["production_price_forecast_method"],
    )
    if isinstance(df_input_data_dayahead, bool) and not df_input_data_dayahead:
        return False
    if "outdoor_temperature_forecast" in input_data_dict["params"]["passed_data"]:
        df_input_data_dayahead["outdoor_temperature_forecast"] = input_data_dict[
            "params"
        ]["passed_data"]["outdoor_temperature_forecast"]
    # The specifics params for the MPC at runtime
    prediction_horizon = input_data_dict["params"]["passed_data"]["prediction_horizon"]
    soc_init = input_data_dict["params"]["passed_data"]["soc_init"]
    soc_final = input_data_dict["params"]["passed_data"]["soc_final"]
    def_total_hours = input_data_dict["params"]["optim_conf"].get(
        "operating_hours_of_each_deferrable_load", None
    )
    def_total_timestep = input_data_dict["params"]["optim_conf"].get(
        "operating_timesteps_of_each_deferrable_load", None
    )
    def_start_timestep = input_data_dict["params"]["optim_conf"][
        "start_timesteps_of_each_deferrable_load"
    ]
    def_end_timestep = input_data_dict["params"]["optim_conf"][
        "end_timesteps_of_each_deferrable_load"
    ]
    opt_res_naive_mpc = input_data_dict["opt"].perform_naive_mpc_optim(
        df_input_data_dayahead,
        input_data_dict["P_PV_forecast"],
        input_data_dict["P_load_forecast"],
        prediction_horizon,
        soc_init,
        soc_final,
        def_total_hours,
        def_total_timestep,
        def_start_timestep,
        def_end_timestep,
    )
    # Save CSV file for publish_data
    if save_data_to_file:
        today = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
        filename = "opt_res_naive_mpc_" + today.strftime("%Y_%m_%d") + ".csv"
    else:  # Just save the latest optimization results
        filename = default_csv_filename
    if not debug:
        opt_res_naive_mpc.to_csv(
            input_data_dict["emhass_conf"]["data_path"] / filename,
            index_label="timestamp",
        )

    if not isinstance(input_data_dict["params"], dict):
        params = json.loads(input_data_dict["params"])
    else:
        params = input_data_dict["params"]

    # if continual_publish, save mpc results to data_path/entities json
    if input_data_dict["retrieve_hass_conf"].get("continual_publish", False) or params[
        "passed_data"
    ].get("entity_save", False):
        # Trigger the publish function, save entity data and not post to HA
        publish_data(input_data_dict, logger, entity_save=True, dont_post=True)

    return opt_res_naive_mpc


def forecast_model_fit(
    input_data_dict: dict, logger: logging.Logger, debug: bool | None = False
) -> tuple[pd.DataFrame, pd.DataFrame, MLForecaster]:
    """Perform a forecast model fit from training data retrieved from Home Assistant.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    :return: The DataFrame containing the forecast data results without and with backtest and the `mlforecaster` object
    :rtype: Tuple[pd.DataFrame, pd.DataFrame, mlforecaster]
    """
    data = copy.deepcopy(input_data_dict["df_input_data"])
    model_type = input_data_dict["params"]["passed_data"]["model_type"]
    var_model = input_data_dict["params"]["passed_data"]["var_model"]
    sklearn_model = input_data_dict["params"]["passed_data"]["sklearn_model"]
    num_lags = input_data_dict["params"]["passed_data"]["num_lags"]
    split_date_delta = input_data_dict["params"]["passed_data"]["split_date_delta"]
    perform_backtest = input_data_dict["params"]["passed_data"]["perform_backtest"]
    # The ML forecaster object
    mlf = MLForecaster(
        data,
        model_type,
        var_model,
        sklearn_model,
        num_lags,
        input_data_dict["emhass_conf"],
        logger,
    )
    # Fit the ML model
    df_pred, df_pred_backtest = mlf.fit(
        split_date_delta=split_date_delta, perform_backtest=perform_backtest
    )
    # Save model
    if not debug:
        filename = model_type + default_pkl_suffix
        filename_path = input_data_dict["emhass_conf"]["data_path"] / filename
        with open(filename_path, "wb") as outp:
            pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
            logger.debug("saved model to " + str(filename_path))
    return df_pred, df_pred_backtest, mlf


def forecast_model_predict(
    input_data_dict: dict,
    logger: logging.Logger,
    use_last_window: bool | None = True,
    debug: bool | None = False,
    mlf: MLForecaster | None = None,
) -> pd.DataFrame:
    r"""Perform a forecast model predict using a previously trained skforecast model.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param use_last_window: True if the 'last_window' option should be used for the \
        custom machine learning forecast model. The 'last_window=True' means that the data \
        that will be used to generate the new forecast will be freshly retrieved from \
        Home Assistant. This data is needed because the forecast model is an auto-regressive \
        model with lags. If 'False' then the data using during the model train is used. Defaults to True
    :type use_last_window: Optional[bool], optional
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    :param mlf: The 'mlforecaster' object previously trained. This is mainly used for debug \
        and unit testing. In production the actual model will be read from a saved pickle file. Defaults to None
    :type mlf: Optional[mlforecaster], optional
    :return: The DataFrame containing the forecast prediction data
    :rtype: pd.DataFrame
    """
    # Load model
    model_type = input_data_dict["params"]["passed_data"]["model_type"]
    filename = model_type + default_pkl_suffix
    filename_path = input_data_dict["emhass_conf"]["data_path"] / filename
    if not debug:
        if filename_path.is_file():
            with open(filename_path, "rb") as inp:
                mlf = pickle.load(inp)
                logger.debug("loaded saved model from " + str(filename_path))
        else:
            logger.error(
                "The ML forecaster file ("
                + str(filename_path)
                + ") was not found, please run a model fit method before this predict method",
            )
            return
    # Make predictions
    if use_last_window:
        data_last_window = copy.deepcopy(input_data_dict["df_input_data"])
    else:
        data_last_window = None
    predictions = mlf.predict(data_last_window)
    # Publish data to a Home Assistant sensor
    model_predict_publish = input_data_dict["params"]["passed_data"][
        "model_predict_publish"
    ]
    model_predict_entity_id = input_data_dict["params"]["passed_data"][
        "model_predict_entity_id"
    ]
    model_predict_device_class = input_data_dict["params"]["passed_data"][
        "model_predict_device_class"
    ]
    model_predict_unit_of_measurement = input_data_dict["params"]["passed_data"][
        "model_predict_unit_of_measurement"
    ]
    model_predict_friendly_name = input_data_dict["params"]["passed_data"][
        "model_predict_friendly_name"
    ]
    publish_prefix = input_data_dict["params"]["passed_data"]["publish_prefix"]
    if model_predict_publish is True:
        # Estimate the current index
        now_precise = datetime.now(
            input_data_dict["retrieve_hass_conf"]["time_zone"]
        ).replace(second=0, microsecond=0)
        if input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "nearest":
            idx_closest = predictions.index.get_indexer(
                [now_precise], method="nearest"
            )[0]
        elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "first":
            idx_closest = predictions.index.get_indexer([now_precise], method="ffill")[
                0
            ]
        elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "last":
            idx_closest = predictions.index.get_indexer([now_precise], method="bfill")[
                0
            ]
        if idx_closest == -1:
            idx_closest = predictions.index.get_indexer(
                [now_precise], method="nearest"
            )[0]
        # Publish Load forecast
        input_data_dict["rh"].post_data(
            predictions,
            idx_closest,
            model_predict_entity_id,
            model_predict_device_class,
            model_predict_unit_of_measurement,
            model_predict_friendly_name,
            type_var="mlforecaster",
            publish_prefix=publish_prefix,
        )
    return predictions


def forecast_model_tune(
    input_data_dict: dict,
    logger: logging.Logger,
    debug: bool | None = False,
    mlf: MLForecaster | None = None,
) -> tuple[pd.DataFrame, MLForecaster]:
    """Tune a forecast model hyperparameters using bayesian optimization.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    :param mlf: The 'mlforecaster' object previously trained. This is mainly used for debug \
        and unit testing. In production the actual model will be read from a saved pickle file. Defaults to None
    :type mlf: Optional[mlforecaster], optional
    :return: The DataFrame containing the forecast data results using the optimized model
    :rtype: pd.DataFrame
    """
    # Load model
    model_type = input_data_dict["params"]["passed_data"]["model_type"]
    filename = model_type + default_pkl_suffix
    filename_path = input_data_dict["emhass_conf"]["data_path"] / filename
    if not debug:
        if filename_path.is_file():
            with open(filename_path, "rb") as inp:
                mlf = pickle.load(inp)
                logger.debug("loaded saved model from " + str(filename_path))
        else:
            logger.error(
                "The ML forecaster file ("
                + str(filename_path)
                + ") was not found, please run a model fit method before this tune method",
            )
            return None, None
    # Tune the model
    df_pred_optim = mlf.tune(debug=debug)
    # Save model
    if not debug:
        filename = model_type + default_pkl_suffix
        filename_path = input_data_dict["emhass_conf"]["data_path"] / filename
        with open(filename_path, "wb") as outp:
            pickle.dump(mlf, outp, pickle.HIGHEST_PROTOCOL)
            logger.debug("Saved model to " + str(filename_path))
    return df_pred_optim, mlf


def regressor_model_fit(
    input_data_dict: dict, logger: logging.Logger, debug: bool | None = False
) -> MLRegressor:
    """Perform a forecast model fit from training data retrieved from Home Assistant.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    """
    data = copy.deepcopy(input_data_dict["df_input_data"])
    if "model_type" in input_data_dict["params"]["passed_data"]:
        model_type = input_data_dict["params"]["passed_data"]["model_type"]
    else:
        logger.error("parameter: 'model_type' not passed")
        return False
    if "regression_model" in input_data_dict["params"]["passed_data"]:
        regression_model = input_data_dict["params"]["passed_data"]["regression_model"]
    else:
        logger.error("parameter: 'regression_model' not passed")
        return False
    if "features" in input_data_dict["params"]["passed_data"]:
        features = input_data_dict["params"]["passed_data"]["features"]
    else:
        logger.error("parameter: 'features' not passed")
        return False
    if "target" in input_data_dict["params"]["passed_data"]:
        target = input_data_dict["params"]["passed_data"]["target"]
    else:
        logger.error("parameter: 'target' not passed")
        return False
    if "timestamp" in input_data_dict["params"]["passed_data"]:
        timestamp = input_data_dict["params"]["passed_data"]["timestamp"]
    else:
        logger.error("parameter: 'timestamp' not passed")
        return False
    if "date_features" in input_data_dict["params"]["passed_data"]:
        date_features = input_data_dict["params"]["passed_data"]["date_features"]
    else:
        logger.error("parameter: 'date_features' not passed")
        return False
    # The MLRegressor object
    mlr = MLRegressor(
        data, model_type, regression_model, features, target, timestamp, logger
    )
    # Fit the ML model
    fit = mlr.fit(date_features=date_features)
    if not fit:
        return False
    # Save model
    if not debug:
        filename = model_type + "_mlr.pkl"
        filename_path = input_data_dict["emhass_conf"]["data_path"] / filename
        with open(filename_path, "wb") as outp:
            pickle.dump(mlr, outp, pickle.HIGHEST_PROTOCOL)
    return mlr


def regressor_model_predict(
    input_data_dict: dict,
    logger: logging.Logger,
    debug: bool | None = False,
    mlr: MLRegressor | None = None,
) -> np.ndarray:
    """Perform a prediction from csv file.

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param debug: True to debug, useful for unit testing, defaults to False
    :type debug: Optional[bool], optional
    """
    if "model_type" in input_data_dict["params"]["passed_data"]:
        model_type = input_data_dict["params"]["passed_data"]["model_type"]
    else:
        logger.error("parameter: 'model_type' not passed")
        return False
    filename = model_type + "_mlr.pkl"
    filename_path = input_data_dict["emhass_conf"]["data_path"] / filename
    if not debug:
        if filename_path.is_file():
            with open(filename_path, "rb") as inp:
                mlr = pickle.load(inp)
        else:
            logger.error(
                "The ML forecaster file was not found, please run a model fit method before this predict method",
            )
            return False
    if "new_values" in input_data_dict["params"]["passed_data"]:
        new_values = input_data_dict["params"]["passed_data"]["new_values"]
    else:
        logger.error("parameter: 'new_values' not passed")
        return False
    # Predict from csv file
    prediction = mlr.predict(new_values)
    mlr_predict_entity_id = input_data_dict["params"]["passed_data"].get(
        "mlr_predict_entity_id", "sensor.mlr_predict"
    )
    mlr_predict_device_class = input_data_dict["params"]["passed_data"].get(
        "mlr_predict_device_class", "power"
    )
    mlr_predict_unit_of_measurement = input_data_dict["params"]["passed_data"].get(
        "mlr_predict_unit_of_measurement", "W"
    )
    mlr_predict_friendly_name = input_data_dict["params"]["passed_data"].get(
        "mlr_predict_friendly_name", "mlr predictor"
    )
    # Publish prediction
    idx = 0
    if not debug:
        input_data_dict["rh"].post_data(
            prediction,
            idx,
            mlr_predict_entity_id,
            mlr_predict_device_class,
            mlr_predict_unit_of_measurement,
            mlr_predict_friendly_name,
            type_var="mlregressor",
        )
    return prediction


def publish_data(
    input_data_dict: dict,
    logger: logging.Logger,
    save_data_to_file: bool | None = False,
    opt_res_latest: pd.DataFrame | None = None,
    entity_save: bool | None = False,
    dont_post: bool | None = False,
) -> pd.DataFrame:
    """
    Publish the data obtained from the optimization results.

    :param input_data_dict:  A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param logger: The passed logger object
    :type logger: logging object
    :param save_data_to_file: If True we will read data from optimization results in dayahead CSV file
    :type save_data_to_file: bool, optional
    :return: The output data of the optimization readed from a CSV file in the data folder
    :rtype: pd.DataFrame
    :param entity_save: Save built entities to data_path/entities
    :type entity_save: bool, optional
    :param dont_post: Do not post to Home Assistant. Works with entity_save
    :type dont_post: bool, optional

    """
    logger.info("Publishing data to HASS instance")
    if input_data_dict:
        if not isinstance(input_data_dict.get("params", {}), dict):
            params = json.loads(input_data_dict["params"])
        else:
            params = input_data_dict.get("params", {})

    # Check if a day ahead optimization has been performed (read CSV file)
    if save_data_to_file:
        today = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
        filename = "opt_res_dayahead_" + today.strftime("%Y_%m_%d") + ".csv"
    # If publish_prefix is passed, check if there is saved entities in data_path/entities with prefix, publish to results
    elif params["passed_data"].get("publish_prefix", "") != "" and not dont_post:
        opt_res_list = []
        opt_res_list_names = []
        publish_prefix = params["passed_data"]["publish_prefix"]
        entity_path = input_data_dict["emhass_conf"]["data_path"] / "entities"
        # Check if items in entity_path
        if os.path.exists(entity_path) and len(os.listdir(entity_path)) > 0:
            # Obtain all files in entity_path
            entity_path_contents = os.listdir(entity_path)
            # Confirm the entity path contains at least one file containing publish prefix or publish_prefix='all'
            if (
                any(publish_prefix in entity for entity in entity_path_contents)
                or publish_prefix == "all"
            ):
                # Loop through all items in entity path
                for entity in entity_path_contents:
                    # If publish_prefix is "all" publish all saved entities to Home Assistant
                    # If publish_prefix matches the prefix from saved entities, publish to Home Assistant
                    if entity != default_metadata_json and (
                        publish_prefix in entity or publish_prefix == "all"
                    ):
                        entity_data = publish_json(
                            entity, input_data_dict, entity_path, logger
                        )
                        if not isinstance(entity_data, bool):
                            opt_res_list.append(entity_data)
                            opt_res_list_names.append(entity.replace(".json", ""))
                        else:
                            return False
                # Build a DataFrame with published entities
                opt_res = pd.concat(opt_res_list, axis=1)
                opt_res.columns = opt_res_list_names
                return opt_res
            else:
                logger.warning(
                    "No saved entity json files that match prefix: "
                    + str(publish_prefix)
                )
                logger.warning("Falling back to opt_res_latest")
        else:
            logger.warning("No saved entity json files in path:" + str(entity_path))
            logger.warning("Falling back to opt_res_latest")
        filename = default_csv_filename
    else:
        filename = default_csv_filename
    if opt_res_latest is None:
        if not os.path.isfile(input_data_dict["emhass_conf"]["data_path"] / filename):
            logger.error("File not found error, run an optimization task first.")
            return
        else:
            opt_res_latest = pd.read_csv(
                input_data_dict["emhass_conf"]["data_path"] / filename,
                index_col="timestamp",
            )
            opt_res_latest.index = pd.to_datetime(opt_res_latest.index)
            opt_res_latest.index.freq = input_data_dict["retrieve_hass_conf"][
                "optimization_time_step"
            ]
    # Estimate the current index
    now_precise = datetime.now(
        input_data_dict["retrieve_hass_conf"]["time_zone"]
    ).replace(second=0, microsecond=0)
    if input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "nearest":
        idx_closest = opt_res_latest.index.get_indexer([now_precise], method="nearest")[
            0
        ]
    elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "first":
        idx_closest = opt_res_latest.index.get_indexer([now_precise], method="ffill")[0]
    elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "last":
        idx_closest = opt_res_latest.index.get_indexer([now_precise], method="bfill")[0]
    if idx_closest == -1:
        idx_closest = opt_res_latest.index.get_indexer([now_precise], method="nearest")[
            0
        ]
    # Publish the data
    publish_prefix = params["passed_data"]["publish_prefix"]
    # Publish PV forecast
    custom_pv_forecast_id = params["passed_data"]["custom_pv_forecast_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["P_PV"],
        idx_closest,
        custom_pv_forecast_id["entity_id"],
        "power",
        custom_pv_forecast_id["unit_of_measurement"],
        custom_pv_forecast_id["friendly_name"],
        type_var="power",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post,
    )
    # Publish Load forecast
    custom_load_forecast_id = params["passed_data"]["custom_load_forecast_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["P_Load"],
        idx_closest,
        custom_load_forecast_id["entity_id"],
        "power",
        custom_load_forecast_id["unit_of_measurement"],
        custom_load_forecast_id["friendly_name"],
        type_var="power",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post,
    )
    cols_published = ["P_PV", "P_Load"]
    # Publish PV curtailment
    if input_data_dict["fcst"].plant_conf["compute_curtailment"]:
        custom_pv_curtailment_id = params["passed_data"]["custom_pv_curtailment_id"]
        input_data_dict["rh"].post_data(
            opt_res_latest["P_PV_curtailment"],
            idx_closest,
            custom_pv_curtailment_id["entity_id"],
            "power",
            custom_pv_curtailment_id["unit_of_measurement"],
            custom_pv_curtailment_id["friendly_name"],
            type_var="power",
            publish_prefix=publish_prefix,
            save_entities=entity_save,
            dont_post=dont_post,
        )
        cols_published = cols_published + ["P_PV_curtailment"]
    # Publish P_hybrid_inverter
    if input_data_dict["fcst"].plant_conf["inverter_is_hybrid"]:
        custom_hybrid_inverter_id = params["passed_data"]["custom_hybrid_inverter_id"]
        input_data_dict["rh"].post_data(
            opt_res_latest["P_hybrid_inverter"],
            idx_closest,
            custom_hybrid_inverter_id["entity_id"],
            "power",
            custom_hybrid_inverter_id["unit_of_measurement"],
            custom_hybrid_inverter_id["friendly_name"],
            type_var="power",
            publish_prefix=publish_prefix,
            save_entities=entity_save,
            dont_post=dont_post,
        )
        cols_published = cols_published + ["P_hybrid_inverter"]
    # Publish deferrable loads
    custom_deferrable_forecast_id = params["passed_data"][
        "custom_deferrable_forecast_id"
    ]
    for k in range(input_data_dict["opt"].optim_conf["number_of_deferrable_loads"]):
        if f"P_deferrable{k}" not in opt_res_latest.columns:
            logger.error(
                f"P_deferrable{k}"
                + " was not found in results DataFrame. Optimization task may need to be relaunched or it did not converge to a solution.",
            )
        else:
            input_data_dict["rh"].post_data(
                opt_res_latest[f"P_deferrable{k}"],
                idx_closest,
                custom_deferrable_forecast_id[k]["entity_id"],
                "power",
                custom_deferrable_forecast_id[k]["unit_of_measurement"],
                custom_deferrable_forecast_id[k]["friendly_name"],
                type_var="deferrable",
                publish_prefix=publish_prefix,
                save_entities=entity_save,
                dont_post=dont_post,
            )
            cols_published = cols_published + [f"P_deferrable{k}"]
    # Publish thermal model data (predicted temperature)
    custom_predicted_temperature_id = params["passed_data"][
        "custom_predicted_temperature_id"
    ]
    for k in range(input_data_dict["opt"].optim_conf["number_of_deferrable_loads"]):
        if "def_load_config" in input_data_dict["opt"].optim_conf.keys():
            if (
                "thermal_config"
                in input_data_dict["opt"].optim_conf["def_load_config"][k]
            ):
                input_data_dict["rh"].post_data(
                    opt_res_latest[f"predicted_temp_heater{k}"],
                    idx_closest,
                    custom_predicted_temperature_id[k]["entity_id"],
                    "temperature",
                    custom_predicted_temperature_id[k]["unit_of_measurement"],
                    custom_predicted_temperature_id[k]["friendly_name"],
                    type_var="temperature",
                    publish_prefix=publish_prefix,
                    save_entities=entity_save,
                    dont_post=dont_post,
                )
                cols_published = cols_published + [f"predicted_temp_heater{k}"]
    # Publish battery power
    if input_data_dict["opt"].optim_conf["set_use_battery"]:
        if "P_batt" not in opt_res_latest.columns:
            logger.error(
                "P_batt was not found in results DataFrame. Optimization task may need to be relaunched or it did not converge to a solution.",
            )
        else:
            custom_batt_forecast_id = params["passed_data"]["custom_batt_forecast_id"]
            input_data_dict["rh"].post_data(
                opt_res_latest["P_batt"],
                idx_closest,
                custom_batt_forecast_id["entity_id"],
                "power",
                custom_batt_forecast_id["unit_of_measurement"],
                custom_batt_forecast_id["friendly_name"],
                type_var="batt",
                publish_prefix=publish_prefix,
                save_entities=entity_save,
                dont_post=dont_post,
            )
            cols_published = cols_published + ["P_batt"]
            custom_batt_soc_forecast_id = params["passed_data"][
                "custom_batt_soc_forecast_id"
            ]
            input_data_dict["rh"].post_data(
                opt_res_latest["SOC_opt"] * 100,
                idx_closest,
                custom_batt_soc_forecast_id["entity_id"],
                "battery",
                custom_batt_soc_forecast_id["unit_of_measurement"],
                custom_batt_soc_forecast_id["friendly_name"],
                type_var="SOC",
                publish_prefix=publish_prefix,
                save_entities=entity_save,
                dont_post=dont_post,
            )
            cols_published = cols_published + ["SOC_opt"]
    # Publish grid power
    custom_grid_forecast_id = params["passed_data"]["custom_grid_forecast_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["P_grid"],
        idx_closest,
        custom_grid_forecast_id["entity_id"],
        "power",
        custom_grid_forecast_id["unit_of_measurement"],
        custom_grid_forecast_id["friendly_name"],
        type_var="power",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post,
    )
    cols_published = cols_published + ["P_grid"]
    # Publish total value of cost function
    custom_cost_fun_id = params["passed_data"]["custom_cost_fun_id"]
    col_cost_fun = [i for i in opt_res_latest.columns if "cost_fun_" in i]
    input_data_dict["rh"].post_data(
        opt_res_latest[col_cost_fun],
        idx_closest,
        custom_cost_fun_id["entity_id"],
        "monetary",
        custom_cost_fun_id["unit_of_measurement"],
        custom_cost_fun_id["friendly_name"],
        type_var="cost_fun",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post,
    )
    # cols_published = cols_published + col_cost_fun
    # Publish the optimization status
    custom_cost_fun_id = params["passed_data"]["custom_optim_status_id"]
    if "optim_status" not in opt_res_latest:
        opt_res_latest["optim_status"] = "Optimal"
        logger.warning(
            "no optim_status in opt_res_latest, run an optimization task first",
        )
    else:
        input_data_dict["rh"].post_data(
            opt_res_latest["optim_status"],
            idx_closest,
            custom_cost_fun_id["entity_id"],
            "",
            custom_cost_fun_id["unit_of_measurement"],
            custom_cost_fun_id["friendly_name"],
            type_var="optim_status",
            publish_prefix=publish_prefix,
            save_entities=entity_save,
            dont_post=dont_post,
        )
        cols_published = cols_published + ["optim_status"]
    # Publish unit_load_cost
    custom_unit_load_cost_id = params["passed_data"]["custom_unit_load_cost_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["unit_load_cost"],
        idx_closest,
        custom_unit_load_cost_id["entity_id"],
        "monetary",
        custom_unit_load_cost_id["unit_of_measurement"],
        custom_unit_load_cost_id["friendly_name"],
        type_var="unit_load_cost",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post,
    )
    cols_published = cols_published + ["unit_load_cost"]
    # Publish unit_prod_price
    custom_unit_prod_price_id = params["passed_data"]["custom_unit_prod_price_id"]
    input_data_dict["rh"].post_data(
        opt_res_latest["unit_prod_price"],
        idx_closest,
        custom_unit_prod_price_id["entity_id"],
        "monetary",
        custom_unit_prod_price_id["unit_of_measurement"],
        custom_unit_prod_price_id["friendly_name"],
        type_var="unit_prod_price",
        publish_prefix=publish_prefix,
        save_entities=entity_save,
        dont_post=dont_post,
    )
    cols_published = cols_published + ["unit_prod_price"]
    # Create a DF resuming what has been published
    opt_res = opt_res_latest[cols_published].loc[[opt_res_latest.index[idx_closest]]]
    return opt_res


def continual_publish(
    input_data_dict: dict, entity_path: pathlib.Path, logger: logging.Logger
):
    """
    If continual_publish is true and a entity file saved in /data_path/entities, continually publish sensor on freq rate, updating entity current state value based on timestamp

    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param entity_path: Path for entities folder in data_path
    :type entity_path: Path
    :param logger: The passed logger object
    :type logger: logging.Logger

    """
    logger.info("Continual publish thread service started")
    freq = input_data_dict["retrieve_hass_conf"].get(
        "optimization_time_step", pd.to_timedelta(1, "minutes")
    )
    entity_path_contents = []
    while True:
        # Sleep for x seconds (using current time as a reference for time left)
        time.sleep(
            max(
                0,
                freq.total_seconds()
                - (
                    datetime.now(
                        input_data_dict["retrieve_hass_conf"]["time_zone"]
                    ).timestamp()
                    % 60
                ),
            )
        )
        # Loop through all saved entity files
        if os.path.exists(entity_path) and len(os.listdir(entity_path)) > 0:
            entity_path_contents = os.listdir(entity_path)
            for entity in entity_path_contents:
                if entity != default_metadata_json:
                    # Call publish_json with entity file, build entity, and publish
                    publish_json(
                        entity,
                        input_data_dict,
                        entity_path,
                        logger,
                        "continual_publish",
                    )
            # Retrieve entity metadata from file
            if os.path.isfile(entity_path / default_metadata_json):
                with open(entity_path / default_metadata_json) as file:
                    metadata = json.load(file)
                    # Check if freq should be shorter
                    if metadata.get("lowest_time_step", None) is not None:
                        freq = pd.to_timedelta(metadata["lowest_time_step"], "minutes")
        pass
    # This function should never return
    return False


def publish_json(
    entity: dict,
    input_data_dict: dict,
    entity_path: pathlib.Path,
    logger: logging.Logger,
    reference: str | None = "",
):
    """
    Extract saved entity data from .json (in data_path/entities), build entity, post results to post_data

    :param entity: json file containing entity data
    :type entity: dict
    :param input_data_dict: A dictionnary with multiple data used by the action functions
    :type input_data_dict: dict
    :param entity_path: Path for entities folder in data_path
    :type entity_path: Path
    :param logger: The passed logger object
    :type logger: logging.Logger
    :param reference: String for identifying who ran the function
    :type reference: str, optional

    """
    # Retrieve entity metadata from file
    if os.path.isfile(entity_path / default_metadata_json):
        with open(entity_path / default_metadata_json) as file:
            metadata = json.load(file)
    else:
        logger.error("unable to located metadata.json in:" + entity_path)
        return False
    # Round current timecode (now)
    now_precise = datetime.now(
        input_data_dict["retrieve_hass_conf"]["time_zone"]
    ).replace(second=0, microsecond=0)
    # Retrieve entity data from file
    entity_data = pd.read_json(entity_path / entity, orient="index")
    # Remove ".json" from string for entity_id
    entity_id = entity.replace(".json", "")
    # Adjust Dataframe from received entity json file
    entity_data.columns = [metadata[entity_id]["name"]]
    entity_data.index.name = "timestamp"
    entity_data.index = pd.to_datetime(entity_data.index).tz_convert(
        input_data_dict["retrieve_hass_conf"]["time_zone"]
    )
    entity_data.index.freq = pd.to_timedelta(
        int(metadata[entity_id]["optimization_time_step"]), "minutes"
    )
    # Calculate the current state value
    if input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "nearest":
        idx_closest = entity_data.index.get_indexer([now_precise], method="nearest")[0]
    elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "first":
        idx_closest = entity_data.index.get_indexer([now_precise], method="ffill")[0]
    elif input_data_dict["retrieve_hass_conf"]["method_ts_round"] == "last":
        idx_closest = entity_data.index.get_indexer([now_precise], method="bfill")[0]
    if idx_closest == -1:
        idx_closest = entity_data.index.get_indexer([now_precise], method="nearest")[0]
    # Call post data
    if reference == "continual_publish":
        logger.debug("Auto Published sensor:")
        logger_levels = "DEBUG"
    else:
        logger_levels = "INFO"
    # post/save entity
    input_data_dict["rh"].post_data(
        data_df=entity_data[metadata[entity_id]["name"]],
        idx=idx_closest,
        entity_id=entity_id,
        device_class=dict.get(metadata[entity_id], "device_class"),
        unit_of_measurement=metadata[entity_id]["unit_of_measurement"],
        friendly_name=metadata[entity_id]["friendly_name"],
        type_var=metadata[entity_id].get("type_var", ""),
        save_entities=False,
        logger_levels=logger_levels,
    )
    return entity_data[metadata[entity_id]["name"]]


def main():
    r"""Define the main command line entry function.

    This function may take several arguments as inputs. You can type `emhass --help` to see the list of options:

    - action: Set the desired action, options are: perfect-optim, dayahead-optim,
      naive-mpc-optim, publish-data, forecast-model-fit, forecast-model-predict, forecast-model-tune

    - config: Define path to the config.yaml file

    - costfun: Define the type of cost function, options are: profit, cost, self-consumption

    - log2file: Define if we should log to a file or not

    - params: Configuration parameters passed from data/options.json if using the add-on

    - runtimeparams: Pass runtime optimization parameters as dictionnary

    - debug: Use True for testing purposes

    """
    # Parsing arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--action",
        type=str,
        help="Set the desired action, options are: perfect-optim, dayahead-optim,\
        naive-mpc-optim, publish-data, forecast-model-fit, forecast-model-predict, forecast-model-tune",
    )
    parser.add_argument(
        "--config", type=str, help="Define path to the config.json/defaults.json file"
    )
    parser.add_argument(
        "--params",
        type=str,
        default=None,
        help="String of configuration parameters passed",
    )
    parser.add_argument(
        "--data", type=str, help="Define path to the Data files (.csv & .pkl)"
    )
    parser.add_argument("--root", type=str, help="Define path emhass root")
    parser.add_argument(
        "--costfun",
        type=str,
        default="profit",
        help="Define the type of cost function, options are: profit, cost, self-consumption",
    )
    parser.add_argument(
        "--log2file",
        type=bool,
        default=False,
        help="Define if we should log to a file or not",
    )
    parser.add_argument(
        "--secrets",
        type=str,
        default=None,
        help="Define secret parameter file (secrets_emhass.yaml) path",
    )
    parser.add_argument(
        "--runtimeparams",
        type=str,
        default=None,
        help="Pass runtime optimization parameters as dictionnary",
    )
    parser.add_argument(
        "--debug",
        type=bool,
        default=False,
        help="Use True for testing purposes",
    )
    args = parser.parse_args()

    # The path to the configuration files
    if args.config is not None:
        config_path = pathlib.Path(args.config)
    else:
        config_path = pathlib.Path(
            str(utils.get_root(__file__, num_parent=3) / "config.json")
        )
    if args.data is not None:
        data_path = pathlib.Path(args.data)
    else:
        data_path = config_path.parent / "data/"
    if args.root is not None:
        root_path = pathlib.Path(args.root)
    else:
        root_path = utils.get_root(__file__, num_parent=1)
    if args.secrets is not None:
        secrets_path = pathlib.Path(args.secrets)
    else:
        secrets_path = pathlib.Path(config_path.parent / "secrets_emhass.yaml")

    associations_path = root_path / "data/associations.csv"
    defaults_path = root_path / "data/config_defaults.json"

    emhass_conf = {}
    emhass_conf["config_path"] = config_path
    emhass_conf["data_path"] = data_path
    emhass_conf["root_path"] = root_path
    emhass_conf["associations_path"] = associations_path
    emhass_conf["defaults_path"] = defaults_path
    # create logger
    logger, ch = utils.get_logger(
        __name__, emhass_conf, save_to_file=bool(args.log2file)
    )

    # Check paths
    logger.debug("config path: " + str(config_path))
    logger.debug("data path: " + str(data_path))
    logger.debug("root path: " + str(root_path))
    if not associations_path.exists():
        logger.error(
            "Could not find associations.csv file in: " + str(associations_path)
        )
        logger.error("Try setting config file path with --associations")
        return False
    if not config_path.exists():
        logger.warning("Could not find config.json file in: " + str(config_path))
        logger.warning("Try setting config file path with --config")
    if not secrets_path.exists():
        logger.warning("Could not find secrets file in: " + str(secrets_path))
        logger.warning("Try setting secrets file path with --secrets")
    if not os.path.isdir(data_path):
        logger.error("Could not find data folder in: " + str(data_path))
        logger.error("Try setting data path with --data")
        return False
    if not os.path.isdir(root_path):
        logger.error("Could not find emhass/src folder in: " + str(root_path))
        logger.error("Try setting emhass root path with --root")
        return False

    # Additional argument
    try:
        parser.add_argument(
            "--version",
            action="version",
            version="%(prog)s " + version("emhass"),
        )
        args = parser.parse_args()
    except Exception:
        logger.info(
            "Version not found for emhass package. Or importlib exited with PackageNotFoundError.",
        )

    # Setup config
    config = {}
    # Check if passed config file is yaml of json, build config accordingly
    if config_path.exists():
        config_file_ending = re.findall(r"(?<=\.).*$", str(config_path))
        if len(config_file_ending) > 0:
            match config_file_ending[0]:
                case "json":
                    config = utils.build_config(
                        emhass_conf, logger, defaults_path, config_path
                    )
                case "yaml":
                    config = utils.build_config(
                        emhass_conf, logger, defaults_path, config_path=config_path
                    )
                case "yml":
                    config = utils.build_config(
                        emhass_conf, logger, defaults_path, config_path=config_path
                    )
    # If unable to find config file, use only defaults_config.json
    else:
        logger.warning(
            "Unable to obtain config.json file, building parameters with only defaults"
        )
        config = utils.build_config(emhass_conf, logger, defaults_path)
    if type(config) is bool and not config:
        raise Exception("Failed to find default config")

    # Obtain secrets from secrets_emhass.yaml?
    params_secrets = {}
    emhass_conf, built_secrets = utils.build_secrets(
        emhass_conf, logger, secrets_path=secrets_path
    )
    params_secrets.update(built_secrets)

    # Build params
    params = utils.build_params(emhass_conf, params_secrets, config, logger)
    if type(params) is bool:
        raise Exception("A error has occurred while building parameters")
    # Add any passed params from args to params
    if args.params:
        params.update(json.loads(args.params))

    input_data_dict = set_input_data_dict(
        emhass_conf,
        args.costfun,
        json.dumps(params),
        args.runtimeparams,
        args.action,
        logger,
        args.debug,
    )
    if type(input_data_dict) is bool:
        raise Exception("A error has occurred while creating action objects")

    # Perform selected action
    if args.action == "perfect-optim":
        opt_res = perfect_forecast_optim(input_data_dict, logger, debug=args.debug)
    elif args.action == "dayahead-optim":
        opt_res = dayahead_forecast_optim(input_data_dict, logger, debug=args.debug)
    elif args.action == "naive-mpc-optim":
        opt_res = naive_mpc_optim(input_data_dict, logger, debug=args.debug)
    elif args.action == "forecast-model-fit":
        df_fit_pred, df_fit_pred_backtest, mlf = forecast_model_fit(
            input_data_dict, logger, debug=args.debug
        )
        opt_res = None
    elif args.action == "forecast-model-predict":
        if args.debug:
            _, _, mlf = forecast_model_fit(input_data_dict, logger, debug=args.debug)
        else:
            mlf = None
        df_pred = forecast_model_predict(
            input_data_dict, logger, debug=args.debug, mlf=mlf
        )
        opt_res = None
    elif args.action == "forecast-model-tune":
        if args.debug:
            _, _, mlf = forecast_model_fit(input_data_dict, logger, debug=args.debug)
        else:
            mlf = None
        df_pred_optim, mlf = forecast_model_tune(
            input_data_dict, logger, debug=args.debug, mlf=mlf
        )
        opt_res = None
    elif args.action == "regressor-model-fit":
        mlr = regressor_model_fit(input_data_dict, logger, debug=args.debug)
        opt_res = None
    elif args.action == "regressor-model-predict":
        if args.debug:
            mlr = regressor_model_fit(input_data_dict, logger, debug=args.debug)
        else:
            mlr = None
        prediction = regressor_model_predict(
            input_data_dict, logger, debug=args.debug, mlr=mlr
        )
        opt_res = None
    elif args.action == "publish-data":
        opt_res = publish_data(input_data_dict, logger)
    else:
        logger.error("The passed action argument is not valid")
        logger.error(
            "Try setting --action: perfect-optim, dayahead-optim, naive-mpc-optim, forecast-model-fit, forecast-model-predict, forecast-model-tune or publish-data"
        )
        opt_res = None
    logger.info(opt_res)
    # Flush the logger
    ch.close()
    logger.removeHandler(ch)
    if (
        args.action == "perfect-optim"
        or args.action == "dayahead-optim"
        or args.action == "naive-mpc-optim"
        or args.action == "publish-data"
    ):
        return opt_res
    elif args.action == "forecast-model-fit":
        return df_fit_pred, df_fit_pred_backtest, mlf
    elif args.action == "forecast-model-predict":
        return df_pred
    elif args.action == "regressor-model-fit":
        return mlr
    elif args.action == "regressor-model-predict":
        return prediction
    elif args.action == "forecast-model-tune":
        return df_pred_optim, mlf
    else:
        return opt_res


if __name__ == "__main__":
    main()



import bz2
import copy
import json
import logging
import os
import pickle
import pickle as cPickle
import re
from datetime import datetime, timedelta
from itertools import zip_longest
from urllib.parse import quote

import numpy as np
import pandas as pd
from pvlib.irradiance import disc
from pvlib.location import Location
from pvlib.modelchain import ModelChain
from pvlib.pvsystem import PVSystem
from pvlib.solarposition import get_solarposition
from pvlib.temperature import TEMPERATURE_MODEL_PARAMETERS
from requests import get
from requests.exceptions import RequestException
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.model_selection import GridSearchCV, TimeSeriesSplit
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from emhass.machine_learning_forecaster import MLForecaster
from emhass.machine_learning_regressor import MLRegressor
from emhass.retrieve_hass import RetrieveHass
from emhass.utils import add_date_features, get_days_list, set_df_index_freq


class Forecast:
    r"""
    Generate weather, load and costs forecasts needed as inputs to the optimization.

    In EMHASS we have basically 4 forecasts to deal with:

    - PV power production forecast (internally based on the weather forecast and the
      characteristics of your PV plant). This is given in Watts.

    - Load power forecast: how much power your house will demand on the next 24h. This
      is given in Watts.

    - PV production selling price forecast: at what price are you selling your excess
      PV production on the next 24h. This is given in EUR/kWh.

    - Load cost forecast: the price of the energy from the grid on the next 24h. This
      is given in EUR/kWh.

    There are methods that are generalized to the 4 forecast needed. For all there
    forecasts it is possible to pass the data either as a passed list of values or by
    reading from a CSV file. With these methods it is then possible to use data from
    external forecast providers.

    Then there are the methods that are specific to each type of forecast and that
    proposed forecast treated and generated internally by this EMHASS forecast class.
    For the weather forecast a first method (`open-meteo`) uses a open-meteos API
    proposing detailed forecasts based on Lat/Lon locations.
    This method seems stable but as with any scrape method it will fail if any changes
    are made to the webpage API. Another method (`solcast`) is using the SolCast PV
    production forecast service. A final method (`solar.forecast`) is using another
    external service: Solar.Forecast, for which just the nominal PV peak installed
    power should be provided. Search the forecast section on the documentation for examples
    on how to implement these different methods.

    The `get_power_from_weather` method is proposed here to convert from irradiance
    data to electrical power. The PVLib module is used to model the PV plant.

    The specific methods for the load forecast are a first method (`naive`) that uses
    a naive approach, also called persistance. It simply assumes that the forecast for
    a future period will be equal to the observed values in a past period. The past
    period is controlled using parameter `delta_forecast`. A second method (`mlforecaster`)
    uses an internal custom forecasting model using machine learning. There is a section
    in the documentation explaining how to use this method.

    .. note:: This custom machine learning model is introduced from v0.4.0. EMHASS \
        proposed this new `mlforecaster` class with `fit`, `predict` and `tune` methods. \
        Only the `predict` method is used here to generate new forecasts, but it is \
        necessary to previously fit a forecaster model and it is a good idea to \
        optimize the model hyperparameters using the `tune` method. See the dedicated \
        section in the documentation for more help.

    For the PV production selling price and Load cost forecasts the privileged method
    is a direct read from a user provided list of values. The list should be passed
    as a runtime parameter during the `curl` to the EMHASS API.

    I reading from a CSV file, it should contain no header and the timestamped data
    should have the following format:

    2021-04-29 00:00:00+00:00,287.07

    2021-04-29 00:30:00+00:00,274.27

    2021-04-29 01:00:00+00:00,243.38

    ...

    The data columns in these files will correspond to the data in the units expected
    for each forecasting method.

    """

    def __init__(
        self,
        retrieve_hass_conf: dict,
        optim_conf: dict,
        plant_conf: dict,
        params: str,
        emhass_conf: dict,
        logger: logging.Logger,
        opt_time_delta: int | None = 24,
        get_data_from_file: bool | None = False,
    ) -> None:
        """
        Define constructor for the forecast class.

        :param retrieve_hass_conf: Dictionary containing the needed configuration
            data from the configuration file, specific to retrieve data from HASS
        :type retrieve_hass_conf: dict
        :param optim_conf: Dictionary containing the needed configuration
            data from the configuration file, specific for the optimization task
        :type optim_conf: dict
        :param plant_conf: Dictionary containing the needed configuration
            data from the configuration file, specific for the modeling of the PV plant
        :type plant_conf: dict
        :param params: Configuration parameters passed from data/options.json
        :type params: str
        :param emhass_conf: Dictionary containing the needed emhass paths
        :type emhass_conf: dict
        :param logger: The passed logger object
        :type logger: logging object
        :param opt_time_delta: The time delta in hours used to generate forecasts,
            a value of 24 will generate 24 hours of forecast data, defaults to 24
        :type opt_time_delta: int, optional
        :param get_data_from_file: Select if data should be retrieved from a
            previously saved pickle useful for testing or directly from connection to
            hass database
        :type get_data_from_file: bool, optional

        """
        self.retrieve_hass_conf = retrieve_hass_conf
        self.optim_conf = optim_conf
        self.plant_conf = plant_conf
        self.freq = self.retrieve_hass_conf["optimization_time_step"]
        self.time_zone = self.retrieve_hass_conf["time_zone"]
        self.method_ts_round = self.retrieve_hass_conf["method_ts_round"]
        self.timeStep = self.freq.seconds / 3600  # in hours
        self.time_delta = pd.to_timedelta(opt_time_delta, "hours")
        self.var_PV = self.retrieve_hass_conf["sensor_power_photovoltaics"]
        self.var_PV_forecast = self.retrieve_hass_conf[
            "sensor_power_photovoltaics_forecast"
        ]
        self.var_load = self.retrieve_hass_conf["sensor_power_load_no_var_loads"]
        self.var_load_new = self.var_load + "_positive"
        self.lat = self.retrieve_hass_conf["Latitude"]
        self.lon = self.retrieve_hass_conf["Longitude"]
        self.emhass_conf = emhass_conf
        self.logger = logger
        self.get_data_from_file = get_data_from_file
        self.var_load_cost = "unit_load_cost"
        self.var_prod_price = "unit_prod_price"
        if (params is None) or (params == "null"):
            self.params = {}
        elif type(params) is dict:
            self.params = params
        else:
            self.params = json.loads(params)

        if self.method_ts_round == "nearest":
            self.start_forecast = pd.Timestamp(
                datetime.now(), tz=self.time_zone
            ).replace(microsecond=0)
        elif self.method_ts_round == "first":
            self.start_forecast = (
                pd.Timestamp(datetime.now(), tz=self.time_zone)
                .replace(microsecond=0)
                .floor(freq=self.freq)
            )
        elif self.method_ts_round == "last":
            self.start_forecast = (
                pd.Timestamp(datetime.now(), tz=self.time_zone)
                .replace(microsecond=0)
                .ceil(freq=self.freq)
            )
        else:
            self.logger.error("Wrong method_ts_round passed parameter")
        # check if weather_forecast_cache, if so get 2x the amount of forecast
        if self.params["passed_data"].get("weather_forecast_cache", False):
            self.end_forecast = (
                self.start_forecast + (self.optim_conf["delta_forecast_daily"] * 2)
            ).replace(microsecond=0)
        else:
            self.end_forecast = (
                self.start_forecast + self.optim_conf["delta_forecast_daily"]
            ).replace(microsecond=0)
        self.forecast_dates = (
            pd.date_range(
                start=self.start_forecast,
                end=self.end_forecast - self.freq,
                freq=self.freq,
                tz=self.time_zone,
            )
            .tz_convert("utc")
            .round(self.freq, ambiguous="infer", nonexistent="shift_forward")
            .tz_convert(self.time_zone)
        )
        if params is not None:
            if "prediction_horizon" in list(self.params["passed_data"].keys()):
                if self.params["passed_data"]["prediction_horizon"] is not None:
                    self.forecast_dates = self.forecast_dates[
                        0 : self.params["passed_data"]["prediction_horizon"]
                    ]

    def get_cached_open_meteo_forecast_json(
        self, max_age: int | None = 30, forecast_days: int = 3
    ) -> dict:
        r"""
        Get weather forecast json from Open-Meteo and cache it for re-use.
        The response json is cached in the local file system and returned
        on subsequent calls until it is older than max_age, at which point
        attempts will be made to replace it with a new version.
        The cached version will not be overwritten until a new version has
        been successfully fetched from Open-Meteo.
        In the event of connectivity issues, the cached version will continue
        to be returned until such time as a new version can be successfully
        fetched from Open-Meteo.
        If you want to force reload, pass max_age value of zero.

        :param max_age: The maximum age of the cached json file, in minutes,
            before it is discarded and a new version fetched from Open-Meteo.
            Defaults to 30 minutes.
        :type max_age: int, optional
        :param forecast_days: The number of days of forecast data required from Open-Meteo.
            One additional day is always fetched from Open-Meteo so there is an extra data in the cache.
            Defaults to 2 days (3 days fetched) to match the prior default.
        :type forecast_days: int, optional
        :return: The json containing the Open-Meteo forecast data
        :rtype: dict

        """

        # Ensure at least 3 weather forecast days (and 1 more than requested)
        if forecast_days is None:
            self.logger.warning(
                "Open-Meteo forecast_days is missing so defaulting to 3 days"
            )
            forecast_days = 3
        elif forecast_days < 3:
            self.logger.warning(
                "Open-Meteo forecast_days is too low (%s) so defaulting to 3 days",
                forecast_days,
            )
            forecast_days = 3
        else:
            forecast_days = forecast_days + 1

        json_path = os.path.abspath(
            self.emhass_conf["data_path"] / "cached-open-meteo-forecast.json"
        )
        # The cached JSON file is always loaded, if it exists, as it is also a fallback
        # in case the REST API call to Open-Meteo fails - the cached JSON will continue to
        # be used until it can successfully fetch a new version from Open-Meteo.
        data = None
        use_cache = False
        if os.path.exists(json_path):
            delta = datetime.now() - datetime.fromtimestamp(os.path.getmtime(json_path))
            json_age = int(delta / timedelta(seconds=60))
            use_cache = json_age < max_age
            self.logger.info(
                "Loading existing cached Open-Meteo JSON file: %s", json_path
            )
            with open(json_path) as json_file:
                data = json.load(json_file)
            if use_cache:
                self.logger.info(
                    "The cached Open-Meteo JSON file is recent (age=%.0fm, max_age=%sm)",
                    json_age,
                    max_age,
                )
            else:
                self.logger.info(
                    "The cached Open-Meteo JSON file is old (age=%.0fm, max_age=%sm)",
                    json_age,
                    max_age,
                )

        if not use_cache:
            self.logger.info("Fetching a new weather forecast from Open-Meteo")
            headers = {"User-Agent": "EMHASS", "Accept": "application/json"}
            url = (
                "https://api.open-meteo.com/v1/forecast?"
                + "latitude="
                + str(round(self.lat, 2))
                + "&longitude="
                + str(round(self.lon, 2))
                + "&minutely_15="
                + "temperature_2m,"
                + "relative_humidity_2m,"
                + "rain,"
                + "cloud_cover,"
                + "wind_speed_10m,"
                + "shortwave_radiation_instant,"
                + "diffuse_radiation_instant,"
                + "direct_normal_irradiance_instant"
                + "&forecast_days="
                + str(forecast_days)
                + "&timezone="
                + quote(str(self.time_zone), safe="")
            )
            try:
                self.logger.debug("Fetching data from Open-Meteo using URL: %s", url)
                response = get(url, headers=headers)
                self.logger.debug("Returned HTTP status code: %s", response.status_code)
                response.raise_for_status()
                """import bz2 # Uncomment to save a serialized data for tests
                import _pickle as cPickle
                with bz2.BZ2File("data/test_response_openmeteo_get_method.pbz2", "w") as f:
                    cPickle.dump(response, f)"""
                data = response.json()
                self.logger.info(
                    "Saving response in Open-Meteo JSON cache file: %s", json_path
                )
                with open(json_path, "w") as json_file:
                    json.dump(response.json(), json_file, indent=2)
            except RequestException:
                self.logger.error(
                    "Failed to fetch weather forecast from Open-Meteo", exc_info=True
                )
                if data is not None:
                    self.logger.warning(
                        "Returning old cached data until next Open-Meteo attempt"
                    )

        return data

    def get_weather_forecast(
        self,
        method: str | None = "open-meteo",
        csv_path: str | None = "data_weather_forecast.csv",
        use_legacy_pvlib: bool | None = False,
    ) -> pd.DataFrame:
        r"""
        Get and generate weather forecast data.

        :param method: The desired method, options are 'open-meteo', 'csv', 'list', 'solcast' and \
            'solar.forecast'. Defaults to 'open-meteo'.
        :type method: str, optional
        :return: The DataFrame containing the forecasted data
        :rtype: pd.DataFrame

        """
        csv_path = self.emhass_conf["data_path"] / csv_path
        w_forecast_cache_path = os.path.abspath(
            self.emhass_conf["data_path"] / "weather_forecast_data.pkl"
        )

        self.logger.info("Retrieving weather forecast data using method = " + method)
        if method == "scrapper":
            self.logger.warning(
                "The scrapper method has been deprecated and the keyword is accepted just for backward compatibility, please change the PV forecast method to open-meteo"
            )
        self.weather_forecast_method = (
            method  # Saving this attribute for later use to identify csv method usage
        )
        if (
            method == "open-meteo" or method == "scrapper"
        ):  # The scrapper option is being left here for backward compatibility
            if not os.path.isfile(w_forecast_cache_path):
                data_raw = self.get_cached_open_meteo_forecast_json(
                    self.optim_conf["open_meteo_cache_max_age"],
                    self.optim_conf["delta_forecast_daily"].days,
                )
                data_15min = pd.DataFrame.from_dict(data_raw["minutely_15"])
                data_15min["time"] = pd.to_datetime(data_15min["time"])
                data_15min.set_index("time", inplace=True)
                data_15min.index = data_15min.index.tz_localize(self.time_zone)

                data_15min = data_15min.rename(
                    columns={
                        "temperature_2m": "temp_air",
                        "relative_humidity_2m": "relative_humidity",
                        "rain": "precipitable_water",
                        "cloud_cover": "cloud_cover",
                        "wind_speed_10m": "wind_speed",
                        "shortwave_radiation_instant": "ghi",
                        "diffuse_radiation_instant": "dhi",
                        "direct_normal_irradiance_instant": "dni",
                    }
                )

                data = data_15min.reindex(self.forecast_dates)
                data.interpolate(
                    method="linear",
                    axis=0,
                    limit=None,
                    limit_direction="both",
                    inplace=True,
                )
                data = set_df_index_freq(data)
                index_utc = data.index.tz_convert("utc")
                index_tz = index_utc.round(
                    freq=data.index.freq, ambiguous="infer", nonexistent="shift_forward"
                ).tz_convert(self.time_zone)
                data.index = index_tz
                data = set_df_index_freq(data)

                # Convert mm to cm and clip the minimum value to 0.1 cm as expected by PVLib
                data["precipitable_water"] = (data["precipitable_water"] / 10).clip(
                    lower=0.1
                )

                if use_legacy_pvlib:
                    # Converting the cloud cover into Global Horizontal Irradiance with a PVLib method
                    data = data.drop(columns=["ghi", "dhi", "dni"])
                    ghi_est = self.cloud_cover_to_irradiance(data["cloud_cover"])
                    data["ghi"] = ghi_est["ghi"]
                    data["dni"] = ghi_est["dni"]
                    data["dhi"] = ghi_est["dhi"]

                # If runtime weather_forecast_cache is true save forecast result to file as cache
                if self.params["passed_data"].get("weather_forecast_cache", False):
                    data = self.set_cached_forecast_data(w_forecast_cache_path, data)
            # Else, open stored weather_forecast_data.pkl file for previous forecast data (cached data)
            # Trim data to match the current required dates
            else:
                data = self.get_cached_forecast_data(w_forecast_cache_path)

        elif method == "solcast":  # using Solcast API
            # Check if weather_forecast_cache is true or if forecast_data file does not exist
            if os.path.isfile(w_forecast_cache_path):
                data = self.get_cached_forecast_data(w_forecast_cache_path)
            # open stored weather_forecast_data.pkl file for previous forecast data (cached data)
            else:
                # Check if weather_forecast_cache_only is true, if so produce error for not finding cache file
                if self.params["passed_data"].get("weather_forecast_cache_only", False):
                    self.logger.error("Unable to obtain Solcast cache file.")
                    self.logger.error(
                        "Try running optimization again with 'weather_forecast_cache_only': false"
                    )
                    self.logger.error(
                        "Optionally, obtain new Solcast cache with runtime parameter 'weather_forecast_cache': true in an optimization, or run the `weather-forecast-cache` action, to pull new data from Solcast and cache."
                    )
                    return False
                else:
                    # Retrieve data from the Solcast API
                    if "solcast_api_key" not in self.retrieve_hass_conf:
                        self.logger.error(
                            "The solcast_api_key parameter was not defined"
                        )
                        return False
                    if "solcast_rooftop_id" not in self.retrieve_hass_conf:
                        self.logger.error(
                            "The solcast_rooftop_id parameter was not defined"
                        )
                        return False
                    headers = {
                        "User-Agent": "EMHASS",
                        "Authorization": "Bearer "
                        + self.retrieve_hass_conf["solcast_api_key"],
                        "content-type": "application/json",
                    }
                    days_solcast = int(
                        len(self.forecast_dates) * self.freq.seconds / 3600
                    )
                    # Split `roof_id` into a list (support comma or space as separator)
                    roof_ids = re.split(
                        r"[,\s]+", self.retrieve_hass_conf["solcast_rooftop_id"].strip()
                    )
                    # Summary list of data
                    total_data_list = [0] * len(self.forecast_dates)
                    # Iteration over individual `roof_id`
                    for roof_id in roof_ids:
                        url = f"https://api.solcast.com.au/rooftop_sites/{roof_id}/forecasts?hours={days_solcast}"
                        response = get(url, headers=headers)
                        """import bz2 # Uncomment to save a serialized data for tests
                        import _pickle as cPickle
                        with bz2.BZ2File("data/test_response_solcast_get_method.pbz2", "w") as f:
                            cPickle.dump(response, f)"""
                        # Verify the request passed
                        if int(response.status_code) == 200:
                            data = response.json()
                        elif (
                            int(response.status_code) == 402
                            or int(response.status_code) == 429
                        ):
                            self.logger.error(
                                "Solcast error: May have exceeded your subscription limit."
                            )
                            return False
                        elif int(response.status_code) >= 400 or (
                            int(response.status_code) >= 202
                            and int(response.status_code) <= 299
                        ):
                            self.logger.error(
                                "Solcast error: There was a issue with the solcast request, check solcast API key and rooftop ID."
                            )
                            self.logger.error(
                                "Solcast error: Check that your subscription is valid and your network can connect to Solcast."
                            )
                            return False
                        # Data processing for the current `roof_id`
                        data_list = []
                        for elm in data["forecasts"]:
                            data_list.append(
                                elm["pv_estimate"] * 1000
                            )  # Converting kW to W
                        # Check if the retrieved data has the correct length
                        if len(data_list) < len(self.forecast_dates):
                            self.logger.error(
                                "Not enough data retrieved from Solcast service, try increasing the time step or use MPC."
                            )
                            return False
                        # Adding the data of the current `roof_id` to the total
                        total_data_list = [
                            total + current
                            for total, current in zip_longest(
                                total_data_list, data_list, fillvalue=0
                            )
                        ]
                    # Trim request results to forecast_dates
                    total_data_list = total_data_list[0 : len(self.forecast_dates)]
                    data_dict = {"ts": self.forecast_dates, "yhat": total_data_list}
                    # Define DataFrame
                    data = pd.DataFrame.from_dict(data_dict)
                    # Define index
                    data.set_index("ts", inplace=True)
                    # If runtime weather_forecast_cache is true save forecast result to file as cache
                    # Trim data to match the current required dates
                    if self.params["passed_data"].get("weather_forecast_cache", False):
                        data = self.set_cached_forecast_data(
                            w_forecast_cache_path, data
                        )

        elif method == "solar.forecast":  # using the solar.forecast API
            # Retrieve data from the solar.forecast API
            if os.path.isfile(w_forecast_cache_path):
                data = self.get_cached_forecast_data(w_forecast_cache_path)
            else:
                if "solar_forecast_kwp" not in self.retrieve_hass_conf:
                    self.logger.warning(
                        "The solar_forecast_kwp parameter was not defined, using dummy values for testing"
                    )
                    self.retrieve_hass_conf["solar_forecast_kwp"] = 5
                if self.retrieve_hass_conf["solar_forecast_kwp"] == 0:
                    self.logger.warning(
                        "The solar_forecast_kwp parameter is set to zero, setting to default 5"
                    )
                    self.retrieve_hass_conf["solar_forecast_kwp"] = 5
                if self.optim_conf["delta_forecast_daily"].days > 1:
                    self.logger.warning(
                        "The free public tier for solar.forecast only provides one day forecasts"
                    )
                    self.logger.warning(
                        "Continuing with just the first day of data, the other days are filled with 0.0."
                    )
                    self.logger.warning(
                        "Use the other available methods for delta_forecast_daily > 1"
                    )
                headers = {"Accept": "application/json"}
                data = pd.DataFrame()
                for i in range(len(self.plant_conf["pv_module_model"])):
                    url = (
                        "https://api.forecast.solar/estimate/"
                        + str(round(self.lat, 2))
                        + "/"
                        + str(round(self.lon, 2))
                        + "/"
                        + str(self.plant_conf["surface_tilt"][i])
                        + "/"
                        + str(self.plant_conf["surface_azimuth"][i] - 180)
                        + "/"
                        + str(self.retrieve_hass_conf["solar_forecast_kwp"])
                    )
                    response = get(url, headers=headers)
                    """import bz2 # Uncomment to save a serialized data for tests
                    import _pickle as cPickle
                    with bz2.BZ2File("data/test_response_solarforecast_get_method.pbz2", "w") as f:
                        cPickle.dump(response.json(), f)"""
                    data_raw = response.json()
                    data_dict = {
                        "ts": list(data_raw["result"]["watts"].keys()),
                        "yhat": list(data_raw["result"]["watts"].values()),
                    }
                    # Form the final DataFrame
                    data_tmp = pd.DataFrame.from_dict(data_dict)
                    data_tmp.set_index("ts", inplace=True)
                    data_tmp.index = pd.to_datetime(data_tmp.index)
                    data_tmp = data_tmp.tz_localize(self.forecast_dates.tz)
                    data_tmp = data_tmp.reindex(index=self.forecast_dates)
                    mask_up_data_df = (
                        data_tmp.copy(deep=True).fillna(method="ffill").isnull()
                    )
                    mask_down_data_df = (
                        data_tmp.copy(deep=True).fillna(method="bfill").isnull()
                    )
                    data_tmp.loc[mask_up_data_df["yhat"], :] = 0.0
                    data_tmp.loc[mask_down_data_df["yhat"], :] = 0.0
                    data_tmp.interpolate(inplace=True, limit=1)
                    data_tmp = data_tmp.fillna(0.0)
                    if len(data) == 0:
                        data = copy.deepcopy(data_tmp)
                    else:
                        data = data + data_tmp
                    # If runtime weather_forecast_cache is true save forecast result to file as cache.
                    # Trim data to match the current required dates
                    if self.params["passed_data"].get("weather_forecast_cache", False):
                        data = self.set_cached_forecast_data(
                            w_forecast_cache_path, data
                        )
        elif method == "csv":  # reading from a csv file
            weather_csv_file_path = csv_path
            # Loading the csv file, we will consider that this is the PV power in W
            data = pd.read_csv(weather_csv_file_path, header=None, names=["ts", "yhat"])
            # Check if the passed data has the correct length
            if len(data) < len(self.forecast_dates):
                self.logger.error("Passed data from CSV is not long enough")
            else:
                # Ensure correct length
                data = data.loc[data.index[0 : len(self.forecast_dates)], :]
                # Define index
                data.index = self.forecast_dates
                data.drop("ts", axis=1, inplace=True)
                data = data.copy().loc[self.forecast_dates]
        elif method == "list":  # reading a list of values
            # Loading data from passed list
            data_list = self.params["passed_data"]["pv_power_forecast"]
            # Check if the passed data has the correct length
            if (
                len(data_list) < len(self.forecast_dates)
                and self.params["passed_data"]["prediction_horizon"] is None
            ):
                self.logger.error("Passed data from passed list is not long enough")
            else:
                # Ensure correct length
                data_list = data_list[0 : len(self.forecast_dates)]
                # Define DataFrame
                data_dict = {"ts": self.forecast_dates, "yhat": data_list}
                data = pd.DataFrame.from_dict(data_dict)
                # Define index
                data.set_index("ts", inplace=True)
        else:
            self.logger.error("Method %r is not valid", method)
            data = None
        self.logger.debug("get_weather_forecast returning:\n%s", data)
        return data

    def cloud_cover_to_irradiance(
        self, cloud_cover: pd.Series, offset: int | None = 35
    ) -> pd.DataFrame:
        """
        Estimates irradiance from cloud cover in the following steps.

        1. Determine clear sky GHI using Ineichen model and
           climatological turbidity.

        2. Estimate cloudy sky GHI using a function of cloud_cover

        3. Estimate cloudy sky DNI using the DISC model.

        4. Calculate DHI from DNI and GHI.

        (This function was copied and modified from PVLib)

        :param cloud_cover: Cloud cover in %.
        :type cloud_cover: pd.Series
        :param offset: Determines the minimum GHI., defaults to 35
        :type offset: Optional[int], optional
        :return: Estimated GHI, DNI, and DHI.
        :rtype: pd.DataFrame
        """
        location = Location(latitude=self.lat, longitude=self.lon)
        solpos = location.get_solarposition(cloud_cover.index)
        cs = location.get_clearsky(
            cloud_cover.index, model="ineichen", solar_position=solpos
        )
        # Using only the linear method
        offset = offset / 100.0
        cloud_cover_unit = copy.deepcopy(cloud_cover) / 100.0
        ghi = (offset + (1 - offset) * (1 - cloud_cover_unit)) * cs["ghi"]
        # Using disc model
        dni = disc(ghi, solpos["zenith"], cloud_cover.index)["dni"]
        dhi = ghi - dni * np.cos(np.radians(solpos["zenith"]))
        irrads = pd.DataFrame({"ghi": ghi, "dni": dni, "dhi": dhi}).fillna(0)
        return irrads

    @staticmethod
    def get_mix_forecast(
        df_now: pd.DataFrame,
        df_forecast: pd.DataFrame,
        alpha: float,
        beta: float,
        col: str,
        ignore_pv_feedback: bool = False,
    ) -> pd.DataFrame:
        """A simple correction method for forecasted data using the current real values of a variable.

        :param df_now: The DataFrame containing the current/real values
        :type df_now: pd.DataFrame
        :param df_forecast: The DataFrame containing the forecast data
        :type df_forecast: pd.DataFrame
        :param alpha: A weight for the forecast data side
        :type alpha: float
        :param beta: A weight for the current/real values sied
        :type beta: float
        :param col: The column variable name
        :type col: str
        :param ignore_pv_feedback: If True, bypass mixing and return original forecast (used during curtailment)
        :type ignore_pv_feedback: bool
        :return: The output DataFrame with the corrected values
        :rtype: pd.DataFrame
        """
        # If ignoring PV feedback (e.g., during curtailment), return original forecast
        if ignore_pv_feedback:
            return df_forecast

        first_fcst = alpha * df_forecast.iloc[0] + beta * df_now[col].iloc[-1]
        df_forecast.iloc[0] = int(round(first_fcst))
        return df_forecast

    def get_power_from_weather(
        self,
        df_weather: pd.DataFrame,
        set_mix_forecast: bool | None = False,
        df_now: pd.DataFrame | None = pd.DataFrame(),
    ) -> pd.Series:
        r"""
        Convert wheater forecast data into electrical power.

        :param df_weather: The DataFrame containing the weather forecasted data. \
            This DF should be generated by the 'get_weather_forecast' method or at \
            least contain the same columns names filled with proper data.
        :type df_weather: pd.DataFrame
        :param set_mix_forecast: Use a mixed forcast strategy to integra now/current values.
        :type set_mix_forecast: Bool, optional
        :param df_now: The DataFrame containing the now/current data.
        :type df_now: pd.DataFrame
        :return: The DataFrame containing the electrical power in Watts
        :rtype: pd.DataFrame

        """
        # If using csv method we consider that yhat is the PV power in W
        if (
            "solar_forecast_kwp" in self.retrieve_hass_conf.keys()
            and self.retrieve_hass_conf["solar_forecast_kwp"] == 0
        ):
            P_PV_forecast = pd.Series(0, index=df_weather.index)
        else:
            if (
                self.weather_forecast_method == "solcast"
                or self.weather_forecast_method == "solar.forecast"
                or self.weather_forecast_method == "csv"
                or self.weather_forecast_method == "list"
            ):
                P_PV_forecast = df_weather["yhat"]
                P_PV_forecast.name = None
            else:  # We will transform the weather data into electrical power
                # Transform to power (Watts)
                # Setting the main parameters of the PV plant
                location = Location(latitude=self.lat, longitude=self.lon)
                temp_params = TEMPERATURE_MODEL_PARAMETERS["sapm"][
                    "close_mount_glass_glass"
                ]
                cec_modules = bz2.BZ2File(
                    self.emhass_conf["root_path"] / "data" / "cec_modules.pbz2", "rb"
                )
                cec_modules = cPickle.load(cec_modules)
                cec_inverters = bz2.BZ2File(
                    self.emhass_conf["root_path"] / "data" / "cec_inverters.pbz2", "rb"
                )
                cec_inverters = cPickle.load(cec_inverters)
                if isinstance(self.plant_conf["pv_module_model"], list):
                    P_PV_forecast = pd.Series(0, index=df_weather.index)
                    for i in range(len(self.plant_conf["pv_module_model"])):
                        # Selecting correct module and inverter
                        module = cec_modules[self.plant_conf["pv_module_model"][i]]
                        inverter = cec_inverters[
                            self.plant_conf["pv_inverter_model"][i]
                        ]
                        # Building the PV system in PVLib
                        system = PVSystem(
                            surface_tilt=self.plant_conf["surface_tilt"][i],
                            surface_azimuth=self.plant_conf["surface_azimuth"][i],
                            module_parameters=module,
                            inverter_parameters=inverter,
                            temperature_model_parameters=temp_params,
                            modules_per_string=self.plant_conf["modules_per_string"][i],
                            strings_per_inverter=self.plant_conf[
                                "strings_per_inverter"
                            ][i],
                        )
                        mc = ModelChain(system, location, aoi_model="physical")
                        # Run the model on the weather DF indexes
                        mc.run_model(df_weather)
                        # Extracting results for AC power
                        P_PV_forecast = P_PV_forecast + mc.results.ac
                else:
                    # Selecting correct module and inverter
                    module = cec_modules[self.plant_conf["pv_module_model"]]
                    inverter = cec_inverters[self.plant_conf["pv_inverter_model"]]
                    # Building the PV system in PVLib
                    system = PVSystem(
                        surface_tilt=self.plant_conf["surface_tilt"],
                        surface_azimuth=self.plant_conf["surface_azimuth"],
                        module_parameters=module,
                        inverter_parameters=inverter,
                        temperature_model_parameters=temp_params,
                        modules_per_string=self.plant_conf["modules_per_string"],
                        strings_per_inverter=self.plant_conf["strings_per_inverter"],
                    )
                    mc = ModelChain(system, location, aoi_model="physical")
                    # Run the model on the weather DF indexes
                    mc.run_model(df_weather)
                    # Extracting results for AC power
                    P_PV_forecast = mc.results.ac
        if set_mix_forecast:
            ignore_pv_feedback = self.params["passed_data"].get("ignore_pv_feedback_during_curtailment", False)
            P_PV_forecast = Forecast.get_mix_forecast(
                df_now,
                P_PV_forecast,
                self.params["passed_data"]["alpha"],
                self.params["passed_data"]["beta"],
                self.var_PV,
                ignore_pv_feedback,
            )
        P_PV_forecast[P_PV_forecast < 0] = 0  # replace any negative PV values with zero
        self.logger.debug("get_power_from_weather returning:\n%s", P_PV_forecast)
        return P_PV_forecast

    @staticmethod
    def compute_solar_angles(
        df: pd.DataFrame, latitude: float, longitude: float
    ) -> pd.DataFrame:
        """
        Compute solar angles (elevation, azimuth) based on timestamps and location.

        :param df: DataFrame with a DateTime index.
        :param latitude: Latitude of the PV system.
        :param longitude: Longitude of the PV system.
        :return: DataFrame with added solar elevation and azimuth.
        """
        df = df.copy()
        solpos = get_solarposition(df.index, latitude, longitude)
        df["solar_elevation"] = solpos["elevation"]
        df["solar_azimuth"] = solpos["azimuth"]
        return df

    def adjust_pv_forecast_data_prep(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Prepare data for adjusting the photovoltaic (PV) forecast.

        This method aligns the actual PV production data with the forecasted data,
        adds additional features for analysis, and separates the predictors (X)
        from the target variable (y).

        :param data: A DataFrame containing the actual PV production data and the
        forecasted PV production data.
        :type data: pd.DataFrame
        """
        # Extract target and predictor
        self.logger.debug("adjust_pv_forecast_data_prep using data:\n%s", data)
        if self.logger.isEnabledFor(logging.DEBUG):
            data.to_csv(
                self.emhass_conf["data_path"]
                / "debug-adjust-pv-forecast-data-prep-input-data.csv"
            )
        P_PV = data[self.var_PV]  # Actual PV production
        P_PV_forecast = data[self.var_PV_forecast]  # Forecasted PV production
        # Define time ranges
        last_day = data.index.max().normalize()  # Last available day
        three_months_ago = last_day - pd.DateOffset(
            days=self.retrieve_hass_conf["historic_days_to_retrieve"]
        )
        # Train/Test: Last historic_days_to_retrieve days (excluding the last day)
        train_test_mask = (data.index >= three_months_ago) & (data.index < last_day)
        self.P_PV_train_test = P_PV[train_test_mask]
        self.P_PV_forecast_train_test = P_PV_forecast[train_test_mask]
        # Validation: Last day only
        validation_mask = data.index >= last_day
        self.P_PV_validation = P_PV[validation_mask]
        self.P_PV_forecast_validation = P_PV_forecast[validation_mask]
        # Ensure data is aligned
        self.data_adjust_pv = pd.concat(
            [P_PV.rename("actual"), P_PV_forecast.rename("forecast")], axis=1
        ).dropna()
        # Add more features
        self.data_adjust_pv = add_date_features(self.data_adjust_pv)
        self.data_adjust_pv = Forecast.compute_solar_angles(
            self.data_adjust_pv, self.lat, self.lon
        )
        # Features (X) and target (y)
        self.X_adjust_pv = self.data_adjust_pv.drop(columns=["actual"])  # Predictors
        self.y_adjust_pv = self.data_adjust_pv["actual"]  # Target: actual PV production
        self.logger.debug(
            "adjust_pv_forecast_data_prep output data:\n%s", self.data_adjust_pv
        )
        if self.logger.isEnabledFor(logging.DEBUG):
            self.data_adjust_pv.to_csv(
                self.emhass_conf["data_path"]
                / "debug-adjust-pv-forecast-data-prep-output-data.csv"
            )

    def adjust_pv_forecast_fit(
        self,
        n_splits: int = 5,
        regression_model: str = "LassoRegression",
        debug: bool | None = False,
    ) -> pd.DataFrame:
        """
        Fit a regression model to adjust the photovoltaic (PV) forecast.

        This method uses historical actual and forecasted PV production data, along with
        additional solar and date features, to train a regression model. The model is
        optimized using a grid search with time-series cross-validation.

        :param n_splits: The number of splits for time-series cross-validation, defaults to 5.
        :type n_splits: int, optional
        :param regression_model: The type of regression model to use. Options include \
            "LassoRegression", "RidgeRegression", etc., defaults to "LassoRegression".
        :type regression_model: str, optional
        :param debug: If True, the model is not saved to disk, useful for debugging, defaults to False.
        :type debug: bool, optional
        :return: A DataFrame containing the adjusted PV forecast.
        :rtype: pd.DataFrame
        """
        # Get regression model and hyperparameter grid
        mlr = MLRegressor(
            self.data_adjust_pv,
            "adjusted_pv_forecast",
            regression_model,
            list(self.X_adjust_pv.columns),
            list(self.y_adjust_pv.name),
            None,
            self.logger,
        )
        base_model, param_grid = mlr.get_regression_model()
        model = make_pipeline(StandardScaler(), base_model)
        # Time-series split
        tscv = TimeSeriesSplit(n_splits=n_splits)
        grid_search = GridSearchCV(
            model, param_grid, cv=tscv, scoring="neg_mean_squared_error", verbose=0
        )
        # Train model
        grid_search.fit(self.X_adjust_pv, self.y_adjust_pv)
        self.model_adjust_pv = grid_search.best_estimator_
        # Calculate training metrics
        y_pred_train = self.model_adjust_pv.predict(self.X_adjust_pv)
        self.rmse = np.sqrt(mean_squared_error(self.y_adjust_pv, y_pred_train))
        self.r2 = r2_score(self.y_adjust_pv, y_pred_train)
        # Log the metrics
        self.logger.info(
            f"PV adjust Training metrics: RMSE = {self.rmse}, R2 = {self.r2}"
        )
        # Save model
        if not debug:
            filename = "adjust_pv_regressor.pkl"
            filename_path = self.emhass_conf["data_path"] / filename
            with open(filename_path, "wb") as outp:
                pickle.dump(self.model_adjust_pv, outp, pickle.HIGHEST_PROTOCOL)

    def adjust_pv_forecast_predict(
        self, forecasted_pv: pd.DataFrame | None = None
    ) -> pd.DataFrame:
        """
        Predict the adjusted photovoltaic (PV) forecast.

        This method uses the trained regression model to predict the adjusted PV forecast
        based on either the validation data stored in `self` or a new forecasted PV data
        passed as input. It applies additional features such as date and solar angles to
        the forecasted PV production data before making predictions. The solar elevation
        is used to avoid negative values and to fix values at the beginning and end of the day.

        :param forecasted_pv: Optional. A DataFrame containing the forecasted PV production data.
                            It must have a DateTime index and a column named "forecast".
                            If not provided, the method will use `self.P_PV_forecast_validation`.
        :type forecasted_pv: pd.DataFrame, optional
        :return: A DataFrame containing the adjusted PV forecast with additional features.
        :rtype: pd.DataFrame
        """
        # Use the provided forecasted PV data or fall back to the validation data in `self`
        if forecasted_pv is not None:
            # Ensure the input DataFrame has the required structure
            if "forecast" not in forecasted_pv.columns:
                raise ValueError(
                    "The input DataFrame must contain a 'forecast' column."
                )
            forecast_data = forecasted_pv.copy()
        else:
            # Use the validation data stored in `self`
            forecast_data = self.P_PV_forecast_validation.rename("forecast").to_frame()
        # Prepare the forecasted PV data
        forecast_data = add_date_features(forecast_data)
        forecast_data = Forecast.compute_solar_angles(forecast_data, self.lat, self.lon)
        # Predict the adjusted forecast
        forecast_data["adjusted_forecast"] = self.model_adjust_pv.predict(forecast_data)

        # Apply solar elevation weighting only for specific cases
        def apply_weighting(row):
            if row["solar_elevation"] <= 0:  # Nighttime or negative solar elevation
                return 0
            elif (
                row["solar_elevation"]
                < self.optim_conf["adjusted_pv_solar_elevation_threshold"]
            ):  # Early morning or late evening
                return max(
                    row["adjusted_forecast"]
                    * (
                        row["solar_elevation"]
                        / self.optim_conf["adjusted_pv_solar_elevation_threshold"]
                    ),
                    0,
                )
            else:  # Daytime with sufficient solar elevation
                return row["adjusted_forecast"]

        forecast_data["adjusted_forecast"] = forecast_data.apply(
            apply_weighting, axis=1
        )
        # If using validation data, calculate validation metrics
        if forecasted_pv is None:
            y_true = self.P_PV_validation.values
            y_pred = forecast_data["adjusted_forecast"].values
            self.validation_rmse = np.sqrt(mean_squared_error(y_true, y_pred))
            self.validation_r2 = r2_score(y_true, y_pred)
            # Log the validation metrics
            self.logger.info(
                f"PV adjust Validation metrics: RMSE = {self.validation_rmse}, R2 = {self.validation_r2}"
            )
        self.logger.debug(
            "adjust_pv_forecast_predict forecast data:\n%s", forecast_data
        )
        if self.logger.isEnabledFor(logging.DEBUG):
            forecast_data.to_csv(
                self.emhass_conf["data_path"]
                / "debug-adjust-pv-forecast-predict-forecast-data.csv"
            )
        # Return the DataFrame with the adjusted forecast
        return forecast_data

    def get_forecast_days_csv(self, timedelta_days: int | None = 1) -> pd.date_range:
        r"""
        Get the date range vector of forecast dates that will be used when loading a CSV file.

        :return: The forecast dates vector
        :rtype: pd.date_range

        """
        start_forecast_csv = pd.Timestamp(datetime.now(), tz=self.time_zone).replace(
            microsecond=0
        )
        if self.method_ts_round == "nearest":
            start_forecast_csv = pd.Timestamp(
                datetime.now(), tz=self.time_zone
            ).replace(microsecond=0)
        elif self.method_ts_round == "first":
            start_forecast_csv = (
                pd.Timestamp(datetime.now(), tz=self.time_zone)
                .replace(microsecond=0)
                .floor(freq=self.freq)
            )
        elif self.method_ts_round == "last":
            start_forecast_csv = (
                pd.Timestamp(datetime.now(), tz=self.time_zone)
                .replace(microsecond=0)
                .ceil(freq=self.freq)
            )
        else:
            self.logger.error("Wrong method_ts_round passed parameter")
        end_forecast_csv = (
            start_forecast_csv + self.optim_conf["delta_forecast_daily"]
        ).replace(microsecond=0)
        forecast_dates_csv = (
            pd.date_range(
                start=start_forecast_csv,
                end=end_forecast_csv + timedelta(days=timedelta_days) - self.freq,
                freq=self.freq,
                tz=self.time_zone,
            )
            .tz_convert("utc")
            .round(self.freq, ambiguous="infer", nonexistent="shift_forward")
            .tz_convert(self.time_zone)
        )
        if self.params is not None:
            if "prediction_horizon" in list(self.params["passed_data"].keys()):
                if self.params["passed_data"]["prediction_horizon"] is not None:
                    forecast_dates_csv = forecast_dates_csv[
                        0 : self.params["passed_data"]["prediction_horizon"]
                    ]
        return forecast_dates_csv

    def get_forecast_out_from_csv_or_list(
        self,
        df_final: pd.DataFrame,
        forecast_dates_csv: pd.date_range,
        csv_path: str,
        data_list: list | None = None,
        list_and_perfect: bool | None = False,
    ) -> pd.DataFrame:
        r"""
        Get the forecast data as a DataFrame from a CSV file.

        The data contained in the CSV file should be a 24h forecast with the same frequency as
        the main 'optimization_time_step' parameter in the configuration file. The timestamp will not be used and
        a new DateTimeIndex is generated to fit the timestamp index of the input data in 'df_final'.

        :param df_final: The DataFrame containing the input data.
        :type df_final: pd.DataFrame
        :param forecast_dates_csv: The forecast dates vector
        :type forecast_dates_csv: pd.date_range
        :param csv_path: The path to the CSV file
        :type csv_path: str
        :return: The data from the CSV file
        :rtype: pd.DataFrame

        """
        if csv_path is None:
            data_dict = {"ts": forecast_dates_csv, "yhat": data_list}
            df_csv = pd.DataFrame.from_dict(data_dict)
            df_csv.index = forecast_dates_csv
            df_csv.drop(["ts"], axis=1, inplace=True)
            df_csv = set_df_index_freq(df_csv)
            if list_and_perfect:
                days_list = df_final.index.day.unique().tolist()
            else:
                days_list = df_csv.index.day.unique().tolist()
        else:
            if not os.path.exists(csv_path):
                csv_path = self.emhass_conf["data_path"] / csv_path
            load_csv_file_path = csv_path
            df_csv = pd.read_csv(load_csv_file_path, header=None, names=["ts", "yhat"])

            first_col = df_csv.iloc[:, 0]
            # If the entire column can be converted to datetime, set it as index
            if pd.to_datetime(first_col, errors="coerce").notna().all():
                df_csv["ts"] = pd.to_datetime(df_csv["ts"], utc=True)
                # Set the timestamp column as the index
                df_csv.set_index("ts", inplace=True)
                df_csv.index = df_csv.index.tz_convert(self.time_zone)
            else:
                df_csv.index = forecast_dates_csv
                df_csv.drop(["ts"], axis=1, inplace=True)
            df_csv = set_df_index_freq(df_csv)
            if list_and_perfect:
                days_list = df_final.index.day.unique().tolist()
            else:
                days_list = df_csv.index.day.unique().tolist()
        forecast_out = pd.DataFrame()
        for day in days_list:
            if csv_path is None:
                df_final = set_df_index_freq(df_final)
                df_tmp = copy.deepcopy(df_final)
            else:
                if list_and_perfect:
                    df_final = set_df_index_freq(df_final)
                    df_tmp = copy.deepcopy(df_final)
                else:
                    df_tmp = copy.deepcopy(df_csv)
            first_elm_index = [i for i, x in enumerate(df_tmp.index.day == day) if x][0]
            last_elm_index = [i for i, x in enumerate(df_tmp.index.day == day) if x][-1]
            fcst_index = pd.date_range(
                start=df_tmp.index[first_elm_index],
                end=df_tmp.index[last_elm_index],
                freq=df_tmp.index.freq,
            )
            first_hour = (
                f"{df_tmp.index[first_elm_index].hour:02d}"
                + ":"
                + f"{df_tmp.index[first_elm_index].minute:02d}"
            )
            last_hour = (
                f"{df_tmp.index[last_elm_index].hour:02d}"
                + ":"
                + f"{df_tmp.index[last_elm_index].minute:02d}"
            )
            if len(forecast_out) == 0:
                if csv_path is None:
                    if list_and_perfect:
                        values_array = df_csv.between_time(first_hour, last_hour).values
                        fcst_index = fcst_index[
                            0 : len(values_array)
                        ]  # Fix for different lengths
                        forecast_out = pd.DataFrame(
                            values_array,
                            index=fcst_index,
                        )
                    else:
                        forecast_out = pd.DataFrame(
                            df_csv.loc[fcst_index, :]
                            .between_time(first_hour, last_hour)
                            .values,
                            index=fcst_index,
                        )
                else:
                    df_csv_filtered_date = df_csv.loc[
                        df_csv.index.strftime("%Y-%m-%d")
                        == fcst_index[0].date().strftime("%Y-%m-%d")
                    ]
                    forecast_out = pd.DataFrame(
                        df_csv_filtered_date.between_time(first_hour, last_hour).values,
                        index=fcst_index,
                    )
            else:
                if csv_path is None:
                    if list_and_perfect:
                        values_array = df_csv.between_time(first_hour, last_hour).values
                        fcst_index = fcst_index[
                            0 : len(values_array)
                        ]  # Fix for different lengths
                        forecast_tp = pd.DataFrame(
                            values_array,
                            index=fcst_index,
                        )
                    else:
                        forecast_tp = pd.DataFrame(
                            df_csv.loc[fcst_index, :]
                            .between_time(first_hour, last_hour)
                            .values,
                            index=fcst_index,
                        )
                else:
                    df_csv_filtered_date = df_csv.loc[
                        df_csv.index.strftime("%Y-%m-%d")
                        == fcst_index[0].date().strftime("%Y-%m-%d")
                    ]
                    forecast_tp = pd.DataFrame(
                        df_csv_filtered_date.between_time(first_hour, last_hour).values,
                        index=fcst_index,
                    )
                forecast_out = pd.concat([forecast_out, forecast_tp], axis=0)
        return forecast_out

    @staticmethod
    def resample_data(data, freq, current_freq):
        r"""
        Resample a DataFrame with a custom frequency.

        :param data: Original time series data with a DateTimeIndex.
        :type data: pd.DataFrame
        :param freq: Desired frequency for resampling (e.g., pd.Timedelta("10min")).
        :type freq: pd.Timedelta
        :return: Resampled data at the specified frequency.
        :rtype: pd.DataFrame
        """
        if freq > current_freq:
            # Downsampling
            # Use 'mean' to aggregate or choose other options ('sum', 'max', etc.)
            resampled_data = data.resample(freq).mean()
        elif freq < current_freq:
            # Upsampling
            # Use 'asfreq' to create empty slots, then interpolate
            resampled_data = data.resample(freq).asfreq()
            resampled_data = resampled_data.interpolate(method="time")
        else:
            # No resampling needed
            resampled_data = data.copy()
        return resampled_data

    @staticmethod
    def get_typical_load_forecast(data, forecast_date):
        r"""
        Forecast the load profile for the next day based on historic data.

        :param data: A DataFrame with a DateTimeIndex containing the historic load data.
                    Must include a 'load' column.
        :type data: pd.DataFrame
        :param forecast_date: The date for which the forecast will be generated.
        :type forecast_date: pd.Timestamp
        :return: A Series with the forecasted load profile for the next day and a list of days used
                to calculate the forecast.
        :rtype: tuple (pd.Series, list)
        """
        # Ensure the 'load' column exists
        if "load" not in data.columns:
            raise ValueError("Data must have a 'load' column.")
        # Filter historic data for the same month and day of the week
        month = forecast_date.month
        day_of_week = forecast_date.dayofweek
        historic_data = data[
            (data.index.month == month) & (data.index.dayofweek == day_of_week)
        ]
        used_days = np.unique(historic_data.index.date)
        # Align all historic data to the forecast day
        aligned_data = []
        for day in used_days:
            daily_data = data[data.index.date == pd.Timestamp(day).date()]
            aligned_daily_data = daily_data.copy()
            aligned_daily_data.index = aligned_daily_data.index.map(
                lambda x: x.replace(
                    year=forecast_date.year,
                    month=forecast_date.month,
                    day=forecast_date.day,
                )
            )
            aligned_data.append(aligned_daily_data)
        # Combine all aligned historic data into a single DataFrame
        combined_data = pd.concat(aligned_data)
        # Compute the mean load for each timestamp
        forecast = combined_data.groupby(combined_data.index).mean()
        return forecast, used_days

    def get_load_forecast(
        self,
        days_min_load_forecast: int | None = 3,
        method: str | None = "typical",
        csv_path: str | None = "data_load_forecast.csv",
        set_mix_forecast: bool | None = False,
        df_now: pd.DataFrame | None = pd.DataFrame(),
        use_last_window: bool | None = True,
        mlf: MLForecaster | None = None,
        debug: bool | None = False,
    ) -> pd.Series:
        r"""
        Get and generate the load forecast data.

        :param days_min_load_forecast: The number of last days to retrieve that \
            will be used to generate a naive forecast, defaults to 3
        :type days_min_load_forecast: int, optional
        :param method: The method to be used to generate load forecast, the options \
            are 'typical' for a typical household load consumption curve, \
            are 'naive' for a persistance model, 'mlforecaster' for using a custom \
            previously fitted machine learning model, 'csv' to read the forecast from \
            a CSV file and 'list' to use data directly passed at runtime as a list of \
            values. Defaults to 'typical'.
        :type method: str, optional
        :param csv_path: The path to the CSV file used when method = 'csv', \
            defaults to "/data/data_load_forecast.csv"
        :type csv_path: str, optional
        :param set_mix_forecast: Use a mixed forcast strategy to integra now/current values.
        :type set_mix_forecast: Bool, optional
        :param df_now: The DataFrame containing the now/current data.
        :type df_now: pd.DataFrame, optional
        :param use_last_window: True if the 'last_window' option should be used for the \
            custom machine learning forecast model. The 'last_window=True' means that the data \
            that will be used to generate the new forecast will be freshly retrieved from \
            Home Assistant. This data is needed because the forecast model is an auto-regressive \
            model with lags. If 'False' then the data using during the model train is used.
        :type use_last_window: Bool, optional
        :param mlf: The 'mlforecaster' object previously trained. This is mainly used for debug \
            and unit testing. In production the actual model will be read from a saved pickle file.
        :type mlf: mlforecaster, optional
        :param debug: The DataFrame containing the now/current data.
        :type debug: Bool, optional
        :return: The DataFrame containing the electrical load power in Watts
        :rtype: pd.DataFrame

        """
        csv_path = self.emhass_conf["data_path"] / csv_path

        if (
            method == "naive" or method == "mlforecaster"
        ):  # retrieving needed data for these methods
            self.logger.info(
                "Retrieving data from hass for load forecast using method = " + method
            )
            var_list = [self.var_load]
            var_replace_zero = None
            var_interp = [self.var_load]
            time_zone_load_foreacast = None
            # We will need to retrieve a new set of load data according to the days_min_load_forecast parameter
            rh = RetrieveHass(
                self.retrieve_hass_conf["hass_url"],
                self.retrieve_hass_conf["long_lived_token"],
                self.freq,
                time_zone_load_foreacast,
                self.params,
                self.emhass_conf,
                self.logger,
            )
            if self.get_data_from_file:
                filename_path = self.emhass_conf["data_path"] / "test_df_final.pkl"
                with open(filename_path, "rb") as inp:
                    rh.df_final, days_list, var_list, rh.ha_config = pickle.load(inp)
                    self.var_load = var_list[0]
                    self.retrieve_hass_conf["sensor_power_load_no_var_loads"] = (
                        self.var_load
                    )
                    var_interp = [var_list[0]]
                    self.var_list = [var_list[0]]
                    rh.var_list = self.var_list
                    self.var_load_new = self.var_load + "_positive"
            else:
                days_list = get_days_list(days_min_load_forecast)
                if not rh.get_data(days_list, var_list):
                    return False
            if not rh.prepare_data(
                self.retrieve_hass_conf["sensor_power_load_no_var_loads"],
                load_negative=self.retrieve_hass_conf["load_negative"],
                set_zero_min=self.retrieve_hass_conf["set_zero_min"],
                var_replace_zero=var_replace_zero,
                var_interp=var_interp,
            ):
                return False
            df = rh.df_final.copy()[[self.var_load_new]]
        if (
            method == "typical"
        ):  # using typical statistical data from a household power consumption
            # Loading data from history file
            model_type = "long_train_data"
            data_path = self.emhass_conf["data_path"] / str(model_type + ".pkl")
            with open(data_path, "rb") as fid:
                data, _, _, _ = pickle.load(fid)
            # Ensure the data index is timezone-aware and matches self.forecast_dates' timezone
            data.index = (
                data.index.tz_localize(self.forecast_dates.tz)
                if data.index.tz is None
                else data.index.tz_convert(self.forecast_dates.tz)
            )
            # Resample the data if needed
            data = data[[self.var_load]]
            current_freq = pd.Timedelta("30min")
            if self.freq != current_freq:
                data = Forecast.resample_data(data, self.freq, current_freq)
            # Generate forecast
            data_list = []
            dates_list = np.unique(self.forecast_dates.date).tolist()
            forecast = pd.DataFrame()
            for date in dates_list:
                forecast_date = pd.Timestamp(date)
                data.columns = ["load"]
                forecast_tmp, used_days = Forecast.get_typical_load_forecast(
                    data, forecast_date
                )
                self.logger.debug(
                    f"Using {len(used_days)} days of data to generate the forecast."
                )
                # Normalize the forecast
                forecast_tmp = (
                    forecast_tmp * self.plant_conf["maximum_power_from_grid"] / 9000
                )
                data_list.extend(forecast_tmp.values.ravel().tolist())
                if len(forecast) == 0:
                    forecast = forecast_tmp
                else:
                    forecast = pd.concat([forecast, forecast_tmp], axis=0)
            forecast_out = forecast.loc[
                forecast.index.intersection(self.forecast_dates)
            ]
            forecast_out.index = self.forecast_dates
            forecast_out.index.name = "ts"
            forecast_out = forecast_out.rename(columns={"load": "yhat"})
        elif method == "naive":  # using a naive approach
            # Old code logic (shifted timestamp problem)
            # mask_forecast_out = (
            #     df.index > days_list[-1] - self.optim_conf["delta_forecast_daily"]
            # )
            # forecast_out = df.copy().loc[mask_forecast_out]
            # forecast_out = forecast_out.rename(columns={self.var_load_new: "yhat"})
            # forecast_out = forecast_out.iloc[0 : len(self.forecast_dates)]
            # forecast_out.index = self.forecast_dates
            # New code logic
            forecast_horizon = len(self.forecast_dates)
            historical_values = df.iloc[-forecast_horizon:]
            forecast_out = pd.DataFrame(
                historical_values.values,
                index=self.forecast_dates,
                columns=["yhat"]
            )
        elif (
            method == "mlforecaster"
        ):  # using a custom forecast model with machine learning
            # Load model
            model_type = self.params["passed_data"]["model_type"]
            filename = model_type + "_mlf.pkl"
            filename_path = self.emhass_conf["data_path"] / filename
            if not debug:
                if filename_path.is_file():
                    with open(filename_path, "rb") as inp:
                        mlf = pickle.load(inp)
                else:
                    self.logger.error(
                        "The ML forecaster file was not found, please run a model fit method before this predict method"
                    )
                    return False
            # Make predictions
            if use_last_window:
                data_last_window = copy.deepcopy(df)
                data_last_window = data_last_window.rename(
                    columns={self.var_load_new: self.var_load}
                )
            else:
                data_last_window = None
            forecast_out = mlf.predict(data_last_window)
            # Force forecast length to avoid mismatches
            self.logger.debug(
                "Number of ML predict forcast data generated (lags_opt): "
                + str(len(forecast_out.index))
            )
            self.logger.debug(
                "Number of forcast dates obtained: " + str(len(self.forecast_dates))
            )
            if len(self.forecast_dates) < len(forecast_out.index):
                forecast_out = forecast_out.iloc[0 : len(self.forecast_dates)]
            # To be removed once bug is fixed
            elif len(self.forecast_dates) > len(forecast_out.index):
                self.logger.error(
                    "Unable to obtain: "
                    + str(len(self.forecast_dates))
                    + " lags_opt values from sensor: power load no var loads, check optimization_time_step/freq and historic_days_to_retrieve/days_to_retrieve parameters"
                )
                return False
            # Define DataFrame
            data_dict = {
                "ts": self.forecast_dates,
                "yhat": forecast_out.values.tolist(),
            }
            data = pd.DataFrame.from_dict(data_dict)
            # Define index
            data.set_index("ts", inplace=True)
            forecast_out = data.copy().loc[self.forecast_dates]
        elif method == "csv":  # reading from a csv file
            load_csv_file_path = csv_path
            df_csv = pd.read_csv(load_csv_file_path, header=None, names=["ts", "yhat"])
            if len(df_csv) < len(self.forecast_dates):
                self.logger.error("Passed data from CSV is not long enough")
            else:
                # Ensure correct length
                df_csv = df_csv.loc[df_csv.index[0 : len(self.forecast_dates)], :]
                # Define index
                df_csv.index = self.forecast_dates
                df_csv.drop(["ts"], axis=1, inplace=True)
                forecast_out = df_csv.copy().loc[self.forecast_dates]
        elif method == "list":  # reading a list of values
            # Loading data from passed list
            data_list = self.params["passed_data"]["load_power_forecast"]
            # Check if the passed data has the correct length
            if (
                len(data_list) < len(self.forecast_dates)
                and self.params["passed_data"]["prediction_horizon"] is None
            ):
                self.logger.error("Passed data from passed list is not long enough")
                return False
            else:
                # Ensure correct length
                data_list = data_list[0 : len(self.forecast_dates)]
                # Define DataFrame
                data_dict = {"ts": self.forecast_dates, "yhat": data_list}
                data = pd.DataFrame.from_dict(data_dict)
                # Define index
                data.set_index("ts", inplace=True)
                forecast_out = data.copy().loc[self.forecast_dates]
        else:
            self.logger.error("Passed method is not valid")
            return False
        P_Load_forecast = copy.deepcopy(forecast_out["yhat"])
        if set_mix_forecast:
            # Load forecasts don't need curtailment protection - always use feedback
            P_Load_forecast = Forecast.get_mix_forecast(
                df_now,
                P_Load_forecast,
                self.params["passed_data"]["alpha"],
                self.params["passed_data"]["beta"],
                self.var_load_new,
                False,  # Never ignore feedback for load forecasts
            )
        self.logger.debug("get_load_forecast returning:\n%s", P_Load_forecast)
        return P_Load_forecast

    def get_load_cost_forecast(
        self,
        df_final: pd.DataFrame,
        method: str | None = "hp_hc_periods",
        csv_path: str | None = "data_load_cost_forecast.csv",
        list_and_perfect: bool | None = False,
    ) -> pd.DataFrame:
        r"""
        Get the unit cost for the load consumption based on multiple tariff \
        periods. This is the cost of the energy from the utility in a vector \
        sampled at the fixed freq value.

        :param df_final: The DataFrame containing the input data.
        :type df_final: pd.DataFrame
        :param method: The method to be used to generate load cost forecast, \
            the options are 'hp_hc_periods' for peak and non-peak hours contracts\
            and 'csv' to load a CSV file, defaults to 'hp_hc_periods'
        :type method: str, optional
        :param csv_path: The path to the CSV file used when method = 'csv', \
            defaults to "data_load_cost_forecast.csv"
        :type csv_path: str, optional
        :return: The input DataFrame with one additionnal column appended containing
            the load cost for each time observation.
        :rtype: pd.DataFrame

        """
        csv_path = self.emhass_conf["data_path"] / csv_path
        if method == "hp_hc_periods":
            df_final[self.var_load_cost] = self.optim_conf["load_offpeak_hours_cost"]
            list_df_hp = []
            for _key, period_hp in self.optim_conf["load_peak_hour_periods"].items():
                list_df_hp.append(
                    df_final[self.var_load_cost].between_time(
                        period_hp[0]["start"], period_hp[1]["end"]
                    )
                )
            for df_hp in list_df_hp:
                df_final.loc[df_hp.index, self.var_load_cost] = self.optim_conf[
                    "load_peak_hours_cost"
                ]
        elif method == "csv":
            forecast_dates_csv = self.get_forecast_days_csv(timedelta_days=0)
            forecast_out = self.get_forecast_out_from_csv_or_list(
                df_final, forecast_dates_csv, csv_path
            )
            # Ensure correct length
            if not list_and_perfect:
                forecast_out = forecast_out[0 : len(self.forecast_dates)]
                df_final = df_final[0 : len(self.forecast_dates)].copy()
            # Convert to Series if needed and align index
            if not isinstance(forecast_out, pd.Series):
                forecast_out = pd.Series(np.ravel(forecast_out), index=df_final.index)
            df_final.loc[:, self.var_load_cost] = forecast_out
        elif method == "list":  # reading a list of values
            # Loading data from passed list
            data_list = self.params["passed_data"]["load_cost_forecast"]
            # Check if the passed data has the correct length
            if (
                len(data_list) < len(self.forecast_dates)
                and self.params["passed_data"]["prediction_horizon"] is None
            ):
                self.logger.error("Passed data from passed list is not long enough")
                return False
            else:
                # Ensure correct length
                data_list = data_list[0 : len(self.forecast_dates)]
                if not list_and_perfect:
                    df_final = df_final.iloc[0 : len(self.forecast_dates)]
                # Define the correct dates
                forecast_dates_csv = self.get_forecast_days_csv(timedelta_days=0)
                forecast_out = self.get_forecast_out_from_csv_or_list(
                    df_final,
                    forecast_dates_csv,
                    None,
                    data_list=data_list,
                    list_and_perfect=list_and_perfect,
                )
                df_final = df_final.copy()
                df_final[self.var_load_cost] = forecast_out
        else:
            self.logger.error("Passed method is not valid")
            return False
        self.logger.debug("get_load_cost_forecast returning:\n%s", df_final)
        return df_final

    def get_prod_price_forecast(
        self,
        df_final: pd.DataFrame,
        method: str | None = "constant",
        csv_path: str | None = "data_prod_price_forecast.csv",
        list_and_perfect: bool | None = False,
    ) -> pd.DataFrame:
        r"""
        Get the unit power production price for the energy injected to the grid.\
        This is the price of the energy injected to the utility in a vector \
        sampled at the fixed freq value.

        :param df_input_data: The DataFrame containing all the input data retrieved
            from hass
        :type df_input_data: pd.DataFrame
        :param method: The method to be used to generate the production price forecast, \
            the options are 'constant' for a fixed constant value and 'csv'\
            to load a CSV file, defaults to 'constant'
        :type method: str, optional
        :param csv_path: The path to the CSV file used when method = 'csv', \
            defaults to "/data/data_load_cost_forecast.csv"
        :type csv_path: str, optional
        :return: The input DataFrame with one additionnal column appended containing
            the power production price for each time observation.
        :rtype: pd.DataFrame

        """
        csv_path = self.emhass_conf["data_path"] / csv_path
        if method == "constant":
            df_final[self.var_prod_price] = self.optim_conf[
                "photovoltaic_production_sell_price"
            ]
        elif method == "csv":
            forecast_dates_csv = self.get_forecast_days_csv(timedelta_days=0)
            forecast_out = self.get_forecast_out_from_csv_or_list(
                df_final, forecast_dates_csv, csv_path
            )
            # Ensure correct length
            if not list_and_perfect:
                forecast_out = forecast_out[0 : len(self.forecast_dates)]
                df_final = df_final[0 : len(self.forecast_dates)].copy()
            # Convert to Series if needed and align index
            if not isinstance(forecast_out, pd.Series):
                forecast_out = pd.Series(np.ravel(forecast_out), index=df_final.index)
            df_final.loc[:, self.var_prod_price] = forecast_out
        elif method == "list":  # reading a list of values
            # Loading data from passed list
            data_list = self.params["passed_data"]["prod_price_forecast"]
            # Check if the passed data has the correct length
            if (
                len(data_list) < len(self.forecast_dates)
                and self.params["passed_data"]["prediction_horizon"] is None
            ):
                self.logger.error("Passed data from passed list is not long enough")
                return False
            else:
                # Ensure correct length
                data_list = data_list[0 : len(self.forecast_dates)]
                if not list_and_perfect:
                    df_final = df_final.iloc[0 : len(self.forecast_dates)]
                # Define the correct dates
                forecast_dates_csv = self.get_forecast_days_csv(timedelta_days=0)
                forecast_out = self.get_forecast_out_from_csv_or_list(
                    df_final,
                    forecast_dates_csv,
                    None,
                    data_list=data_list,
                    list_and_perfect=list_and_perfect,
                )
                df_final = df_final.copy()
                df_final[self.var_prod_price] = forecast_out
        else:
            self.logger.error("Passed method is not valid")
            return False
        self.logger.debug("get_prod_price_forecast returning:\n%s", df_final)
        return df_final

    def get_cached_forecast_data(self, w_forecast_cache_path) -> pd.DataFrame:
        r"""
        Get cached weather forecast data from file.

        :param w_forecast_cache_path: the path to file.
        :type method: Any
        :return: The DataFrame containing the forecasted data
        :rtype: pd.DataFrame

        """
        with open(w_forecast_cache_path, "rb") as file:
            data = cPickle.load(file)
            if not isinstance(data, pd.DataFrame) or len(data) < len(
                self.forecast_dates
            ):
                self.logger.error(
                    "There has been a error obtaining cached forecast data."
                )
                self.logger.error(
                    "Try running optimization again with 'weather_forecast_cache': true, or run action `weather-forecast-cache`, to pull new data from forecast API and cache."
                )
                self.logger.warning(
                    "Removing old forecast cache file. Next optimization will pull data from forecast API, unless 'weather_forecast_cache_only': true"
                )
                os.remove(w_forecast_cache_path)
                return False
            # Filter cached forecast data to match current forecast_dates start-end range (reduce forecast Dataframe size to appropriate length)
            if (
                self.forecast_dates[0] in data.index
                and self.forecast_dates[-1] in data.index
            ):
                data = data.loc[self.forecast_dates[0] : self.forecast_dates[-1]]
                self.logger.info(
                    "Retrieved forecast data from the previously saved cache."
                )
            else:
                self.logger.error(
                    "Unable to obtain cached forecast data within the requested timeframe range."
                )
                self.logger.error(
                    "Try running optimization again (not using cache). Optionally, add runtime parameter 'weather_forecast_cache': true to pull new data from forecast API and cache."
                )
                self.logger.warning(
                    "Removing old forecast cache file. Next optimization will pull data from forecast API, unless 'weather_forecast_cache_only': true"
                )
                os.remove(w_forecast_cache_path)
                return False
            return data

    def set_cached_forecast_data(self, w_forecast_cache_path, data) -> pd.DataFrame:
        r"""
        Set generated weather forecast data to file.
        Trim data to match the original requested forecast dates

        :param w_forecast_cache_path: the path to file.
        :type method: Any
        :param: The DataFrame containing the forecasted data
        :type: pd.DataFrame
        :return: The DataFrame containing the forecasted data
        :rtype: pd.DataFrame

        """
        with open(w_forecast_cache_path, "wb") as file:
            cPickle.dump(data, file)
            if not os.path.isfile(w_forecast_cache_path):
                self.logger.warning("forecast data could not be saved to file.")
            else:
                self.logger.info(
                    "Saved the forecast results to cache, for later reference."
                )

        # Trim cached data to match requested dates
        end_forecast = (
            self.start_forecast + self.optim_conf["delta_forecast_daily"]
        ).replace(microsecond=0)
        forecast_dates = (
            pd.date_range(
                start=self.start_forecast,
                end=end_forecast - self.freq,
                freq=self.freq,
                tz=self.time_zone,
            )
            .tz_convert("utc")
            .round(self.freq, ambiguous="infer", nonexistent="shift_forward")
            .tz_convert(self.time_zone)
        )
        data = data.loc[forecast_dates[0] : forecast_dates[-1]]
        return data
